# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2016-2023 Sören Gebbert and mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""
Global configuration
"""

import ast
import configparser
import csv
import os
from json import load as json_load

__license__ = "GPLv3"
__author__ = "Sören Gebbert, Anika Weinmann"
__copyright__ = (
    "Copyright 2016-2023, Sören Gebbert and mundialis GmbH & Co. KG"
)
__maintainer__ = "mundialis GmbH & Co. KG"

if os.environ.get("DEFAULT_CONFIG_PATH"):
    DEFAULT_CONFIG_PATH = os.environ["DEFAULT_CONFIG_PATH"]
else:
    DEFAULT_CONFIG_PATH = "/etc/default/actinia"
if not os.path.isfile(DEFAULT_CONFIG_PATH):
    open(DEFAULT_CONFIG_PATH, "a").close()

allow_list = [
    "d.legend",
    "d.rast",
    "d.rast.multi",
    "d.vect",
    "exporter",
    "g.findfile",
    "g.gisenv",
    "g.list",
    "g.mapset",
    "g.proj",
    "g.region",
    "g.remove",
    "g.rename",
    "g.version",
    "i.atcorr",
    "i.cluster",
    "i.colors.enhance",
    "i.gensig",
    "i.group",
    "i.landsat.toar",
    "i.maxlik",
    "i.pansharpen",
    "i.segment",
    "i.tasscap",
    "i.vi",
    "importer",
    "r.blend",
    "r.buffer",
    "r.buffer.lowmem",
    "r.carve",
    "r.category",
    "r.circle",
    "r.clump",
    "r.coin",
    "r.colors",
    "r.colors.out",
    "r.composite",
    "r.compress",
    "r.contour",
    "r.cost",
    "r.covar",
    "r.cross",
    "r.describe",
    "r.distance",
    "r.drain",
    "r.fill.dir",
    "r.fillnulls",
    "r.flow",
    "r.grow",
    "r.grow.distance",
    "r.gwflow",
    "r.his",
    "r.horizon",
    "r.info",
    "r.kappa",
    "r.lake",
    "r.latlong",
    "r.li.cwed",
    "r.li.dominance",
    "r.li.edgedensity",
    "r.li.mpa",
    "r.li.mps",
    "r.li.padcv",
    "r.li.padrange",
    "r.li.padsd",
    "r.li.patchdensity",
    "r.li.patchnum",
    "r.li.pielou",
    "r.li.renyi",
    "r.li.richness",
    "r.li.shannon",
    "r.li.shape",
    "r.li.simpson",
    "r.mapcalc",
    "r.mask",
    "r.mfilter",
    "r.mode",
    "r.neighbors",
    "r.null",
    "r.out.png",
    "r.param.scale",
    "r.patch",
    "r.plane",
    "r.profile",
    "r.proj",
    "r.quant",
    "r.quantile",
    "r.random",
    "r.random.cells",
    "r.random.surface",
    "r.reclass",
    "r.reclass.area",
    "r.recode",
    "r.region",
    "r.regression.line",
    "r.regression.multi",
    "r.regression.series",
    "r.relief",
    "r.report",
    "r.resamp.bspline",
    "r.resamp.filter",
    "r.resamp.interp",
    "r.resamp.rst",
    "r.resamp.stats",
    "r.resample",
    "r.rescale",
    "r.rescale.eq",
    "r.rgb",
    "r.ros",
    "r.series",
    "r.series.accumulate",
    "r.series.interp",
    "r.shade",
    "r.sim.sediment",
    "r.sim.water",
    "r.slope.aspect",
    "r.solute.transport",
    "r.spread",
    "r.spreadpath",
    "r.statistics",
    "r.stats",
    "r.stats.quantile",
    "r.stats.zonal",
    "r.stream.extract",
    "r.sun",
    "r.sunhours",
    "r.sunmask",
    "r.support",
    "r.support.stats",
    "r.surf.area",
    "r.surf.contour",
    "r.surf.fractal",
    "r.surf.gauss",
    "r.surf.idw",
    "r.surf.random",
    "r.terraflow",
    "r.texture",
    "r.thin",
    "r.tile",
    "r.tileset",
    "r.timestamp",
    "r.to.rast3",
    "r.to.rast3elev",
    "r.to.vect",
    "r.topidx",
    "r.topmodel",
    "r.transect",
    "r.univar",
    "r.uslek",
    "r.usler",
    "r.viewshed",
    "r.vol.dem",
    "r.volume",
    "r.walk",
    "r.water.outlet",
    "r.watershed",
    "r.what",
    "r.what.color",
    "t.create",
    "t.info",
    "t.list",
    "t.rast.accdetect",
    "t.rast.accumulate",
    "t.rast.aggr_func",
    "t.rast.aggregate",
    "t.rast.aggregate.ds",
    "t.rast.algebra",
    "t.rast.colors",
    "t.rast.extract",
    "t.rast.gapfill",
    "t.rast.list",
    "t.rast.mapcalc",
    "t.rast.sample",
    "t.rast.series",
    "t.rast.univar",
    "t.rast.what",
    "t.register",
    "t.remove",
    "t.unregister",
    "v.buffer",
    "v.clean",
    "v.db.select",
    "v.db.univar",
    "v.db.update",
    "v.dissolve",
    "v.in.ascii",
    "v.overlay",
    "v.patch",
    "v.random",
    "v.rast.stats",
    "v.select",
    "v.what.rast",
    "v.what.strds",
]
# GRASS ADDON: maybe not installed:
# * r.regression.series
# * r.vol.dem
# * d.rast.multi (https://github.com/mundialis/d_rast_multi)
# * t.rast.aggr_func (https://github.com/mundialis/openeo-addons/tree/master/
#           t.rast.udf)
# * t.rast.sample (https://github.com/mundialis/t.rast.sample)


class Configuration(object):
    def __init__(self):
        """
        The constructor creates default parameter that can be overwritten with
        a read() call
        """

        home = os.getenv("HOME")

        """GRASS"""
        # The GRASS global database
        # The GRASS database in which the user locations
        # are stored. This is the base path, der user group
        # will be added on runtime
        self.GRASS_DATABASE = "%s/actinia/grassdb" % home
        # The default GRASS location in the global database that
        # is used for location generation
        self.GRASS_USER_DATABASE = "%s/actinia/userdata" % home
        # The directory to store temporary GRASS databases
        self.GRASS_DEFAULT_LOCATION = "nc_spm_08"
        # Directory to store exported resources
        self.GRASS_TMP_DATABASE = "%s/actinia/workspace/temp_db" % home
        self.GRASS_RESOURCE_DIR = "%s/actinia/resources" % home
        # The size quota of the resource storage in Gigibit
        self.GRASS_RESOURCE_QUOTA = 100
        # Installation directory of GRASS
        self.GRASS_GIS_BASE = "/usr/local/grass/"
        self.GRASS_GIS_START_SCRIPT = "/usr/local/bin/grass"
        self.GRASS_ADDON_PATH = "%s/.grass8/addons/" % home
        self.GRASS_MODULES_XML_PATH = os.path.join(
            self.GRASS_GIS_BASE, "gui", "wxpython", "xml", "module_items.xml"
        )
        # The path to the activation script of the python2 venv (old)
        self.GRASS_VENV = (
            "%s/src/actinia/grass_venv/bin/activate_this.py" % home
        )
        # ADDITIONAL_ALLOWED_MODULES: modules to extend MODULE_ALLOW_LIST
        # e.g. ['i.cutlines', 'r.learn.ml2']
        self.ADDITIONAL_ALLOWED_MODULES = []

        """
        LIMITS
        DEFAULT LIMITS when a user is created
        """
        # Maximum number of cells that are allowed to process
        self.MAX_CELL_LIMIT = 1000 * 1000 * 1000
        # Maximum number of seconds an async process is allowed to run
        self.PROCESS_TIME_LIMT = 600
        # Maximum number of processes in a process chain
        self.PROCESS_NUM_LIMIT = 1000
        # The number of queues that process jobs
        self.NUMBER_OF_WORKERS = 3

        """
        API SETTINGS
        """
        # CHECK_CREDENTIALS: If set False all credential checks are disabled
        self.CHECK_CREDENTIALS = True
        # CHECK_LIMITS: If set False all limit checks are disabled
        self.CHECK_LIMITS = True
        # LOG_API_CALL: If set False the API calls are not logged
        self.LOG_API_CALL = True
        # LOGIN_REQUIRED: If set False, login is not required
        self.LOGIN_REQUIRED = True
        # FORCE_HTTPS_URLS: Force the use of https in response urls that
        # point to actinia services like status URL or data storage
        self.FORCE_HTTPS_URLS = False
        # PLUGINS: e.g. ["actinia_satellite_plugin","actinia_statistic_plugin"]
        self.PLUGINS = []
        # ENDPOINTS_CONFIG: configuration csv file for endpoints
        self.ENDPOINTS_CONFIG = None
        # AUTHENTICATION: If set False no authentication is needed
        self.AUTHENTICATION = True

        """
        KEYCLOAK: has only to be set if keycloak server is configured with
        actinia client and actinia attributes for the users
        """
        # Json file generated by keycloak with configuration,
        # e.g. /etc/default/keycloak.json
        self.KEYCLOAK_CONFIG_PATH = None
        # e.g. /actinia-user/
        self.KEYCLOAK_GROUP_PREFIX = None
        # Prefix to distinguish parameters inside keycloak from parameters
        # used by other applications, e.g. actinia
        self.KEYCLOAK_ATTR_PREFIX = None
        # KEYCLOAK parameter read from json configured in KEYCLOAK_CONFIG_PATH
        self.KEYCLOAK_URL = None
        self.KEYCLOAK_CLIENT_ID = None
        self.KEYCLOAK_REALM = None
        self.KEYCLOAK_CLIENT_SECRET_KEY = None

        """
        REDIS
        """
        # The hostname of the redis server
        self.REDIS_SERVER_URL = "127.0.0.1"
        # The port of the redis server
        self.REDIS_SERVER_PORT = 6379
        # The password of the redis server
        self.REDIS_SERVER_PW = None
        # Default expire time is 10 days for resource logs, that are used for
        # calculating the price of resource usage
        self.REDIS_RESOURCE_EXPIRE_TIME = 864000
        # The hostname of the redis work queue server
        self.REDIS_QUEUE_SERVER_URL = "127.0.0.1"
        # The port of the redis work queue server
        self.REDIS_QUEUE_SERVER_PORT = 6379
        # The password of the redis work queue server
        self.REDIS_QUEUE_SERVER_PASSWORD = None
        # This is the time the rq:job will be stored in the redis
        self.REDIS_QUEUE_JOB_TTL = None
        # The base name of the redis worker queue, it will be extended by a
        # numerical suffix that represents the worker id/number database to
        # re-queue it, usually this is not necessary
        self.WORKER_QUEUE_PREFIX = "job_queue"
        # The base name of the redis worker queue logfile, it will be extended
        # by a numerical suffix that represents the worker id/number
        self.WORKER_LOGFILE = "%s/actinia/workspace/tmp/worker.log" % home

        """
        QUEUE
        """
        # The number of queues that process jobs
        self.NUMBER_OF_WORKERS = 3
        # The hostname of the redis work queue server
        self.REDIS_QUEUE_SERVER_URL = "127.0.0.1"
        # The port of the redis work queue server
        self.REDIS_QUEUE_SERVER_PORT = 6379
        # The password of the redis work queue server
        self.REDIS_QUEUE_SERVER_PASSWORD = None
        # This is the time the rq:job will be stored in redis
        self.REDIS_QUEUE_JOB_TTL = None
        # The prefix for the name of the redis worker queue.
        # It will be extended by a numerical suffix that represents
        # the worker id/number database to re-queue it, usually this is not
        # necessary. If QUEUE_TYPE = per_job, it is extended by the
        # resource_id of the job, if QUEUE_TYPE = per_user, it is extended
        # by the user_id
        self.WORKER_QUEUE_PREFIX = "job_queue"
        # Type of queue.
        # "local":  Single queue for all jobs, processed by same actinia
        #           instance via multiprocessing.
        # "redis":  Number of queues is equal to number of workers as set
        #           in config NUMBER_OF_WORKERS, processed by different
        #           actinia instances (actinia worker).
        # "per_job": Separate queue for each job, config for NUMBER_OF_WORKERS
        #           is ignored. Processed by different actinia instance
        #           (actinia worker). Resource_id will be added to above
        #           WORKER_QUEUE_PREFIX.
        # "per_user": Separate queue for each user, config for NUMBER_OF_WORKERS
        #           is ignored. Processed by different actinia instance
        #           (actinia worker). User_id will be added to above
        #           WORKER_QUEUE_PREFIX.
        # future ideas
        # - redis separate queue per process type
        # - redis separate queue per resource consumption
        self.QUEUE_TYPE = "local"
        # Separate configuration for queue_type for synchronous requests which
        # might not want to be queued.
        self.QUEUE_TYPE_OVERWRITE = "local"

        """
        MISC
        """
        # The directory to store temporary files
        self.TMP_WORKDIR = "%s/actinia/workspace/tmp" % home
        self.SECRET_KEY = (
            "This is a very secret key that is used to sign tokens"
        )
        # The directory to cache downloaded data
        self.DOWNLOAD_CACHE = "/tmp/download_cache"
        # The quota of the download cache in Gigabit
        self.DOWNLOAD_CACHE_QUOTA = 100
        # If True the interim results (temporary mapset) are saved
        self.SAVE_INTERIM_RESULTS = False
        self.SAVE_INTERIM_RESULTS_ENDPOINTS_CFG = None
        self.INTERIM_SAVING_ENDPOINTS = {
            "AsyncEphemeralResource".lower(): "AsyncEphemeralResource",
            "AsyncEphemeralExportResource".lower(): "AsyncEphemeralExportResource",
            "AsyncPersistentResource".lower(): "AsyncPersistentResource",
        }
        self.INCLUDE_ADDITIONAL_MAPSET_PATTERN = None

        """
        LOGGING
        """
        # Logging: 1 Error, 2 Warning, 3 Info, 4 Debug
        self.LOG_LEVEL = 1
        # The logging interface to use: "stdout" or "fluentd" (file will
        # always be used)
        self.LOG_INTERFACE = "stdout"
        # The logformat to use for stdout: "colored" or "json"
        self.LOG_STDOUT_FORMAT = "colored"
        # The logformat to use in file: "colored" or "json"
        self.LOG_FILE_FORMAT = "colored"
        # The logformat for the stderr logger: "plain" or "default", default
        # will use STDOUT / FILE format
        self.LOG_STDERR_FORMAT = "plain"
        # The Fluentd host used for fluent logging
        self.LOG_FLUENT_HOST = "127.0.0.1"
        # The Fluentd host used for fluent logging
        self.LOG_FLUENT_PORT = 24224

        """
        MANAGEMENT
        """
        # If authentication is not required, a default user is used
        self.DEFAULT_USER = "user"
        # If authentication is not required, a default group is used
        self.DEFAULT_USER_GROUP = "group"

        # Not in config file
        self.MODULE_ALLOW_LIST = (
            allow_list  # The default allow list of GRASS modules
        )

        # AWS S3 ADMIN CREDENTIALS
        self.S3_AWS_ACCESS_KEY_ID = ""
        self.S3_AWS_SECRET_ACCESS_KEY = ""
        self.S3_AWS_DEFAULT_REGION = ""
        # The AWS S3 bucket to store user resources
        self.S3_AWS_RESOURCE_BUCKET = ""

        # GOOGLE CLOUD STORAGE (GCS) CREDENTIALS
        # This file stores the Google Cloud Storage credentials
        self.GOOGLE_APPLICATION_CREDENTIALS = "/etc/GCS_credentials.json"
        # The google project
        self.GOOGLE_CLOUD_PROJECT = ""
        # The Google Cloud Storage bucket to store user resources
        self.GCS_RESOURCE_BUCKET = ""

        """
        WEBHOOK
        """
        # Webhook finished retry
        self.WEBHOOK_RETRIES = 6
        self.WEBHOOK_SLEEP = 10

    def __str__(self):
        string = ""
        for entry in dir(self):
            if "__" not in entry and entry in self.__dict__:
                string += "%s=%s\n" % (entry, str(self.__dict__[entry]))

        return string

    def read_keycloak_config(self, key_cloak_config_path=None):
        """Read keycloak configuration json"""
        if key_cloak_config_path is None:
            key_cloak_config_path = self.KEYCLOAK_CONFIG_PATH
        if os.path.isfile(key_cloak_config_path):
            with open(key_cloak_config_path) as f:
                keycloak_cfg = json_load(f)
                self.KEYCLOAK_URL = keycloak_cfg["auth-server-url"]
                self.KEYCLOAK_REALM = keycloak_cfg["realm"]
                self.KEYCLOAK_CLIENT_ID = keycloak_cfg["resource"]
                self.KEYCLOAK_CLIENT_SECRET_KEY = keycloak_cfg["credentials"][
                    "secret"
                ]
        else:
            raise Exception(
                "KEYCLOAK_CONFIG_PATH is not a valid keycloak configuration "
                "for actinia"
            )

    def write(self, path=DEFAULT_CONFIG_PATH):
        """Save the configuration into a file

        Args:
            path (str): The path to the configuration file

        Raises:
            IOError: If unable to write config file

        """
        config = configparser.ConfigParser()

        config.add_section("GRASS")

        config.set("GRASS", "GRASS_DATABASE", self.GRASS_DATABASE)
        config.set("GRASS", "GRASS_USER_DATABASE", self.GRASS_USER_DATABASE)
        config.set(
            "GRASS", "GRASS_DEFAULT_LOCATION", self.GRASS_DEFAULT_LOCATION
        )
        config.set("GRASS", "GRASS_TMP_DATABASE", self.GRASS_TMP_DATABASE)
        config.set("GRASS", "GRASS_RESOURCE_DIR", self.GRASS_RESOURCE_DIR)
        config.set(
            "GRASS", "GRASS_RESOURCE_QUOTA", str(self.GRASS_RESOURCE_QUOTA)
        )
        config.set("GRASS", "GRASS_GIS_BASE", self.GRASS_GIS_BASE)
        config.set(
            "GRASS", "GRASS_GIS_START_SCRIPT", self.GRASS_GIS_START_SCRIPT
        )
        config.set("GRASS", "GRASS_ADDON_PATH", self.GRASS_ADDON_PATH)
        config.set(
            "GRASS", "GRASS_MODULES_XML_PATH", self.GRASS_MODULES_XML_PATH
        )
        config.set("GRASS", "GRASS_VENV", self.GRASS_VENV)

        config.add_section("LIMITS")
        config.set("LIMITS", "MAX_CELL_LIMIT", str(self.MAX_CELL_LIMIT))
        config.set("LIMITS", "PROCESS_TIME_LIMT", str(self.PROCESS_TIME_LIMT))
        config.set("LIMITS", "PROCESS_NUM_LIMIT", str(self.PROCESS_NUM_LIMIT))

        config.add_section("API")
        config.set("API", "CHECK_CREDENTIALS", str(self.CHECK_CREDENTIALS))
        config.set("API", "CHECK_LIMITS", str(self.CHECK_LIMITS))
        config.set("API", "LOG_API_CALL", str(self.LOG_API_CALL))
        config.set("API", "LOGIN_REQUIRED", str(self.LOGIN_REQUIRED))
        config.set("API", "FORCE_HTTPS_URLS", str(self.FORCE_HTTPS_URLS))
        config.set("API", "PLUGINS", str(self.PLUGINS))
        config.set("API", "ENDPOINTS_CONFIG", str(self.ENDPOINTS_CONFIG))
        config.set("API", "AUTHENTICATION", str(self.AUTHENTICATION))

        config.add_section("KEYCLOAK")
        config.set(
            "KEYCLOAK",
            "KEYCLOAK_CONFIG_PATH",
            str(self.KEYCLOAK_CONFIG_PATH),
        )
        config.set(
            "KEYCLOAK",
            "KEYCLOAK_GROUP_PREFIX",
            str(self.KEYCLOAK_GROUP_PREFIX),
        )
        config.set(
            "KEYCLOAK",
            "KEYCLOAK_ATTR_PREFIX",
            str(self.KEYCLOAK_ATTR_PREFIX),
        )
        config.set(
            "KEYCLOAK",
            "KEYCLOAK_URL",
            str(self.KEYCLOAK_URL),
        )
        config.set(
            "KEYCLOAK",
            "KEYCLOAK_CLIENT_ID",
            str(self.KEYCLOAK_CLIENT_ID),
        )
        config.set(
            "KEYCLOAK",
            "KEYCLOAK_REALM",
            str(self.KEYCLOAK_REALM),
        )
        config.set(
            "KEYCLOAK",
            "KEYCLOAK_CLIENT_SECRET_KEY",
            str(self.KEYCLOAK_CLIENT_SECRET_KEY),
        )

        config.add_section("REDIS")
        config.set("REDIS", "REDIS_SERVER_URL", self.REDIS_SERVER_URL)
        config.set("REDIS", "REDIS_SERVER_PORT", str(self.REDIS_SERVER_PORT))
        config.set("REDIS", "REDIS_SERVER_PW", str(self.REDIS_SERVER_PW))
        config.set(
            "REDIS",
            "REDIS_RESOURCE_EXPIRE_TIME",
            str(self.REDIS_RESOURCE_EXPIRE_TIME),
        )
        config.set("REDIS", "WORKER_LOGFILE", str(self.WORKER_LOGFILE))

        config.add_section("QUEUE")
        config.set("QUEUE", "NUMBER_OF_WORKERS", str(self.NUMBER_OF_WORKERS))
        config.set(
            "QUEUE", "REDIS_QUEUE_SERVER_URL", self.REDIS_QUEUE_SERVER_URL
        )
        config.set(
            "QUEUE",
            "REDIS_QUEUE_SERVER_PORT",
            str(self.REDIS_QUEUE_SERVER_PORT),
        )
        config.set(
            "QUEUE",
            "REDIS_QUEUE_SERVER_PASSWORD",
            str(self.REDIS_QUEUE_SERVER_PASSWORD),
        )
        config.set(
            "QUEUE", "REDIS_QUEUE_JOB_TTL", str(self.REDIS_QUEUE_JOB_TTL)
        )
        config.set(
            "QUEUE", "WORKER_QUEUE_PREFIX", str(self.WORKER_QUEUE_PREFIX)
        )
        config.set("QUEUE", "QUEUE_TYPE", self.QUEUE_TYPE)
        config.set("QUEUE", "QUEUE_TYPE_OVERWRITE", self.QUEUE_TYPE_OVERWRITE)

        config.add_section("MISC")
        config.set("MISC", "DOWNLOAD_CACHE", self.DOWNLOAD_CACHE)
        config.set(
            "MISC", "DOWNLOAD_CACHE_QUOTA", str(self.DOWNLOAD_CACHE_QUOTA)
        )
        config.set("MISC", "TMP_WORKDIR", self.TMP_WORKDIR)
        config.set("MISC", "SECRET_KEY", self.SECRET_KEY)
        config.set(
            "MISC", "SAVE_INTERIM_RESULTS", str(self.SAVE_INTERIM_RESULTS)
        )
        if self.SAVE_INTERIM_RESULTS_ENDPOINTS_CFG:
            config.set(
                "MISC",
                "SAVE_INTERIM_RESULTS_ENDPOINTS_CFG",
                self.SAVE_INTERIM_RESULTS_ENDPOINTS_CFG,
            )
        config.set(
            "MISC",
            "INTERIM_SAVING_ENDPOINTS",
            str(self.INTERIM_SAVING_ENDPOINTS),
        )
        config.set(
            "MISC",
            "INCLUDE_ADDITIONAL_MAPSET_PATTERN",
            str(self.INCLUDE_ADDITIONAL_MAPSET_PATTERN),
        )

        config.add_section("LOGGING")
        config.set("LOGGING", "LOG_INTERFACE", self.LOG_INTERFACE)
        config.set("LOGGING", "LOG_STDOUT_FORMAT", self.LOG_STDOUT_FORMAT)
        config.set("LOGGING", "LOG_FILE_FORMAT", self.LOG_FILE_FORMAT)
        config.set("LOGGING", "LOG_STDERR_FORMAT", self.LOG_STDERR_FORMAT)
        config.set("LOGGING", "LOG_FLUENT_HOST", str(self.LOG_FLUENT_HOST))
        config.set("LOGGING", "LOG_FLUENT_PORT", str(self.LOG_FLUENT_PORT))
        config.set("LOGGING", "LOG_LEVEL", str(self.LOG_LEVEL))

        config.add_section("MANAGEMENT")
        config.set("MANAGEMENT", "DEFAULT_USER", str(self.DEFAULT_USER))
        config.set(
            "MANAGEMENT", "DEFAULT_USER_GROUP", str(self.DEFAULT_USER_GROUP)
        )

        config.add_section("AWS_S3")
        config.set("AWS_S3", "S3_AWS_ACCESS_KEY_ID", self.S3_AWS_ACCESS_KEY_ID)
        config.set(
            "AWS_S3", "S3_AWS_SECRET_ACCESS_KEY", self.S3_AWS_SECRET_ACCESS_KEY
        )
        config.set(
            "AWS_S3", "S3_AWS_DEFAULT_REGION", self.S3_AWS_DEFAULT_REGION
        )
        config.set(
            "AWS_S3", "S3_AWS_RESOURCE_BUCKET", self.S3_AWS_RESOURCE_BUCKET
        )

        config.add_section("GCS")
        config.set(
            "GCS",
            "GOOGLE_APPLICATION_CREDENTIALS",
            self.GOOGLE_APPLICATION_CREDENTIALS,
        )
        config.set("GCS", "GCS_RESOURCE_BUCKET", self.GCS_RESOURCE_BUCKET)
        config.set("GCS", "GOOGLE_CLOUD_PROJECT", self.GOOGLE_CLOUD_PROJECT)

        config.add_section("WEBHOOK")
        config.set("WEBHOOK", "WEBHOOK_RETRIES", str(self.WEBHOOK_RETRIES))
        config.set("WEBHOOK", "WEBHOOK_SLEEP", str(self.WEBHOOK_SLEEP))

        with open(path, "w") as configfile:
            config.write(configfile)

    def read(self, path=DEFAULT_CONFIG_PATH):
        """Read the configuration from a file

        Args:
            path (str): The path to the configuration file

        Raises:
            IOError: If unable to read config file

        """
        config = configparser.ConfigParser()
        with open(path, "r") as configfile:
            config.read_file(configfile)

            if config.has_section("GRASS"):
                if config.has_option("GRASS", "GRASS_DATABASE"):
                    self.GRASS_DATABASE = config.get("GRASS", "GRASS_DATABASE")
                if config.has_option("GRASS", "GRASS_USER_DATABASE"):
                    self.GRASS_USER_DATABASE = config.get(
                        "GRASS", "GRASS_USER_DATABASE"
                    )
                if config.has_option("GRASS", "GRASS_DEFAULT_LOCATION"):
                    self.GRASS_DEFAULT_LOCATION = config.get(
                        "GRASS", "GRASS_DEFAULT_LOCATION"
                    )
                if config.has_option("GRASS", "GRASS_TMP_DATABASE"):
                    self.GRASS_TMP_DATABASE = config.get(
                        "GRASS", "GRASS_TMP_DATABASE"
                    )
                if config.has_option("GRASS", "GRASS_RESOURCE_DIR"):
                    self.GRASS_RESOURCE_DIR = config.get(
                        "GRASS", "GRASS_RESOURCE_DIR"
                    )
                if config.has_option("GRASS", "GRASS_RESOURCE_QUOTA"):
                    self.GRASS_RESOURCE_QUOTA = config.getint(
                        "GRASS", "GRASS_RESOURCE_QUOTA"
                    )
                if config.has_option("GRASS", "GRASS_GIS_BASE"):
                    self.GRASS_GIS_BASE = config.get("GRASS", "GRASS_GIS_BASE")
                if config.has_option("GRASS", "GRASS_GIS_START_SCRIPT"):
                    self.GRASS_GIS_START_SCRIPT = config.get(
                        "GRASS", "GRASS_GIS_START_SCRIPT"
                    )
                if config.has_option("GRASS", "GRASS_ADDON_PATH"):
                    self.GRASS_ADDON_PATH = config.get(
                        "GRASS", "GRASS_ADDON_PATH"
                    )
                if config.has_option("GRASS", "GRASS_MODULES_XML_PATH"):
                    self.GRASS_MODULES_XML_PATH = config.get(
                        "GRASS", "GRASS_MODULES_XML_PATH"
                    )
                if config.has_option("GRASS", "GRASS_VENV"):
                    self.GRASS_VENV = config.get("GRASS", "GRASS_VENV")
                if config.has_option(
                    "MANAGEMENT", "ADDITIONAL_ALLOWED_MODULES"
                ):
                    self.ADDITIONAL_ALLOWED_MODULES = ast.literal_eval(
                        config.get("MANAGEMENT", "ADDITIONAL_ALLOWED_MODULES")
                    )
                    self.MODULE_ALLOW_LIST.extend(
                        self.ADDITIONAL_ALLOWED_MODULES
                    )
                    self.MODULE_ALLOW_LIST = list(set(self.MODULE_ALLOW_LIST))

            if config.has_section("LIMITS"):
                if config.has_option("LIMITS", "MAX_CELL_LIMIT"):
                    self.MAX_CELL_LIMIT = config.getint(
                        "LIMITS", "MAX_CELL_LIMIT"
                    )
                if config.has_option("LIMITS", "PROCESS_TIME_LIMT"):
                    self.PROCESS_TIME_LIMT = config.getint(
                        "LIMITS", "PROCESS_TIME_LIMT"
                    )
                if config.has_option("LIMITS", "PROCESS_NUM_LIMIT"):
                    self.PROCESS_NUM_LIMIT = config.getint(
                        "LIMITS", "PROCESS_NUM_LIMIT"
                    )

            if config.has_section("API"):
                if config.has_option("API", "CHECK_CREDENTIALS"):
                    self.CHECK_CREDENTIALS = config.getboolean(
                        "API", "CHECK_CREDENTIALS"
                    )
                if config.has_option("API", "CHECK_LIMITS"):
                    self.CHECK_LIMITS = config.getboolean(
                        "API", "CHECK_LIMITS"
                    )
                if config.has_option("API", "LOG_API_CALL"):
                    self.LOG_API_CALL = config.getboolean(
                        "API", "LOG_API_CALL"
                    )
                if config.has_option("API", "LOGIN_REQUIRED"):
                    self.LOGIN_REQUIRED = config.getboolean(
                        "API", "LOGIN_REQUIRED"
                    )
                if config.has_option("API", "FORCE_HTTPS_URLS"):
                    self.FORCE_HTTPS_URLS = config.getboolean(
                        "API", "FORCE_HTTPS_URLS"
                    )
                if config.has_option("API", "PLUGINS"):
                    self.PLUGINS = ast.literal_eval(
                        config.get("API", "PLUGINS")
                    )
                if config.has_option("API", "ENDPOINTS_CONFIG"):
                    self.ENDPOINTS_CONFIG = config.get(
                        "API", "ENDPOINTS_CONFIG"
                    )
                if config.has_option("API", "AUTHENTICATION"):
                    self.AUTHENTICATION = config.getboolean(
                        "API", "AUTHENTICATION"
                    )

            if config.has_section("KEYCLOAK"):
                if config.has_option("KEYCLOAK", "CONFIG_PATH"):
                    keycloak_cfg_path = config.get("KEYCLOAK", "CONFIG_PATH")
                    if os.path.isfile(keycloak_cfg_path):
                        self.KEYCLOAK_CONFIG_PATH = keycloak_cfg_path
                        self.read_keycloak_config()
                    else:
                        print(
                            "Keycloak is configured, but configfile is not "
                            "an existing file! Using Redis for user "
                            "management."
                        )
                if config.has_option("KEYCLOAK", "GROUP_PREFIX"):
                    self.KEYCLOAK_GROUP_PREFIX = config.get(
                        "KEYCLOAK", "GROUP_PREFIX"
                    )
                if config.has_option("KEYCLOAK", "ATTR_PREFIX"):
                    self.KEYCLOAK_ATTR_PREFIX = config.get(
                        "KEYCLOAK", "ATTR_PREFIX"
                    )

            if config.has_section("REDIS"):
                if config.has_option("REDIS", "REDIS_SERVER_URL"):
                    self.REDIS_SERVER_URL = config.get(
                        "REDIS", "REDIS_SERVER_URL"
                    )
                if config.has_option("REDIS", "REDIS_SERVER_PORT"):
                    self.REDIS_SERVER_PORT = config.getint(
                        "REDIS", "REDIS_SERVER_PORT"
                    )
                if config.has_option("REDIS", "REDIS_SERVER_PW"):
                    self.REDIS_SERVER_PW = config.get(
                        "REDIS", "REDIS_SERVER_PW"
                    )
                if config.has_option("REDIS", "REDIS_RESOURCE_EXPIRE_TIME"):
                    self.REDIS_RESOURCE_EXPIRE_TIME = config.getint(
                        "REDIS", "REDIS_RESOURCE_EXPIRE_TIME"
                    )
                if config.has_option("REDIS", "WORKER_LOGFILE"):
                    self.WORKER_LOGFILE = config.get("REDIS", "WORKER_LOGFILE")

            if config.has_section("QUEUE"):
                if config.has_option("QUEUE", "NUMBER_OF_WORKERS"):
                    self.NUMBER_OF_WORKERS = config.getint(
                        "QUEUE", "NUMBER_OF_WORKERS"
                    )
                if config.has_option("QUEUE", "REDIS_QUEUE_SERVER_URL"):
                    self.REDIS_QUEUE_SERVER_URL = config.get(
                        "QUEUE", "REDIS_QUEUE_SERVER_URL"
                    )
                if config.has_option("QUEUE", "REDIS_QUEUE_SERVER_PORT"):
                    self.REDIS_QUEUE_SERVER_PORT = config.get(
                        "QUEUE", "REDIS_QUEUE_SERVER_PORT"
                    )
                if config.has_option("QUEUE", "REDIS_QUEUE_SERVER_PASSWORD"):
                    self.REDIS_QUEUE_SERVER_PASSWORD = config.get(
                        "QUEUE", "REDIS_QUEUE_SERVER_PASSWORD"
                    )
                if config.has_option("QUEUE", "REDIS_QUEUE_JOB_TTL"):
                    self.REDIS_QUEUE_JOB_TTL = config.get(
                        "QUEUE", "REDIS_QUEUE_JOB_TTL"
                    )
                if config.has_option("QUEUE", "WORKER_QUEUE_PREFIX"):
                    self.WORKER_QUEUE_PREFIX = config.get(
                        "QUEUE", "WORKER_QUEUE_PREFIX"
                    )
                if config.has_option("QUEUE", "QUEUE_TYPE"):
                    self.QUEUE_TYPE = config.get("QUEUE", "QUEUE_TYPE")
                if config.has_option("QUEUE", "QUEUE_TYPE_OVERWRITE"):
                    self.QUEUE_TYPE_OVERWRITE = config.get(
                        "QUEUE", "QUEUE_TYPE_OVERWRITE"
                    )

            if config.has_section("MISC"):
                if config.has_option("MISC", "DOWNLOAD_CACHE"):
                    self.DOWNLOAD_CACHE = config.get("MISC", "DOWNLOAD_CACHE")
                if config.has_option("MISC", "DOWNLOAD_CACHE_QUOTA"):
                    self.DOWNLOAD_CACHE_QUOTA = config.getint(
                        "MISC", "DOWNLOAD_CACHE_QUOTA"
                    )
                if config.has_option("MISC", "TMP_WORKDIR"):
                    self.TMP_WORKDIR = config.get("MISC", "TMP_WORKDIR")
                if config.has_option("MISC", "SECRET_KEY"):
                    self.SECRET_KEY = config.get("MISC", "SECRET_KEY")
                if config.has_option("MISC", "LOG_LEVEL"):
                    self.LOG_LEVEL = config.getint("MISC", "LOG_LEVEL")
                if config.has_option("MISC", "SAVE_INTERIM_RESULTS"):
                    save_interim = config.get("MISC", "SAVE_INTERIM_RESULTS")
                    if save_interim == "True":
                        self.SAVE_INTERIM_RESULTS = True
                    elif save_interim == "False":
                        self.SAVE_INTERIM_RESULTS = False
                    else:
                        self.SAVE_INTERIM_RESULTS = save_interim
                if config.has_option(
                    "MISC", "SAVE_INTERIM_RESULTS_ENDPOINTS_CFG"
                ):
                    cfg = config.get(
                        "MISC", "SAVE_INTERIM_RESULTS_ENDPOINTS_CFG"
                    )
                    if os.path.isfile(cfg):
                        self.SAVE_INTERIM_RESULTS_ENDPOINTS_CFG = cfg
                        with open(cfg, mode="r") as inp:
                            reader = csv.reader(inp, delimiter=";")
                            endpoints_dict = {
                                row[0].lower(): row[1]
                                for row in reader
                                if len(row) == 2
                            }
                            self.INTERIM_SAVING_ENDPOINTS.update(
                                endpoints_dict
                            )
                if config.has_option(
                    "MISC", "INCLUDE_ADDITIONAL_MAPSET_PATTERN"
                ):
                    self.INCLUDE_ADDITIONAL_MAPSET_PATTERN = config.get(
                        "MISC", "INCLUDE_ADDITIONAL_MAPSET_PATTERN"
                    )

            if config.has_section("LOGGING"):
                if config.has_option("LOGGING", "LOG_INTERFACE"):
                    self.LOG_INTERFACE = config.get("LOGGING", "LOG_INTERFACE")
                if config.has_option("LOGGING", "LOG_STDOUT_FORMAT"):
                    self.LOG_STDOUT_FORMAT = config.get(
                        "LOGGING", "LOG_STDOUT_FORMAT"
                    )
                if config.has_option("LOGGING", "LOG_FILE_FORMAT"):
                    self.LOG_FILE_FORMAT = config.get(
                        "LOGGING", "LOG_FILE_FORMAT"
                    )
                if config.has_option("LOGGING", "LOG_STDERR_FORMAT"):
                    self.LOG_STDERR_FORMAT = config.get(
                        "LOGGING", "LOG_STDERR_FORMAT"
                    )
                if config.has_option("LOGGING", "LOG_FLUENT_HOST"):
                    self.LOG_FLUENT_HOST = config.get(
                        "LOGGING", "LOG_FLUENT_HOST"
                    )
                if config.has_option("LOGGING", "LOG_FLUENT_PORT"):
                    self.LOG_FLUENT_PORT = config.getint(
                        "LOGGING", "LOG_FLUENT_PORT"
                    )
                if config.has_option("LOGGING", "LOG_LEVEL"):
                    self.LOG_LEVEL = config.getint("LOGGING", "LOG_LEVEL")

            if config.has_section("MANAGEMENT"):
                if config.has_option("MANAGEMENT", "DEFAULT_USER"):
                    self.DEFAULT_USER = config.get(
                        "MANAGEMENT", "DEFAULT_USER"
                    )
                if config.has_option("MANAGEMENT", "DEFAULT_USER_GROUP"):
                    self.DEFAULT_USER_GROUP = config.get(
                        "MANAGEMENT", "DEFAULT_USER_GROUP"
                    )

            if config.has_section("GCS"):
                if config.has_option("GCS", "GOOGLE_APPLICATION_CREDENTIALS"):
                    self.GOOGLE_APPLICATION_CREDENTIALS = config.get(
                        "GCS", "GOOGLE_APPLICATION_CREDENTIALS"
                    )
                if config.has_option("GCS", "GCS_RESOURCE_BUCKET"):
                    self.GCS_RESOURCE_BUCKET = config.get(
                        "GCS", "GCS_RESOURCE_BUCKET"
                    )
                if config.has_option("GCS", "GOOGLE_CLOUD_PROJECT"):
                    self.GOOGLE_CLOUD_PROJECT = config.get(
                        "GCS", "GOOGLE_CLOUD_PROJECT"
                    )

            if config.has_section("AWS_S3"):
                if config.has_option("AWS_S3", "S3_AWS_ACCESS_KEY_ID"):
                    self.S3_AWS_ACCESS_KEY_ID = config.get(
                        "AWS_S3", "S3_AWS_ACCESS_KEY_ID"
                    )
                if config.has_option("AWS_S3", "S3_AWS_SECRET_ACCESS_KEY"):
                    self.S3_AWS_SECRET_ACCESS_KEY = config.get(
                        "AWS_S3", "S3_AWS_SECRET_ACCESS_KEY"
                    )
                if config.has_option("AWS_S3", "S3_AWS_DEFAULT_REGION"):
                    self.S3_AWS_DEFAULT_REGION = config.get(
                        "AWS_S3", "S3_AWS_DEFAULT_REGION"
                    )
                if config.has_option("AWS_S3", "S3_AWS_RESOURCE_BUCKET"):
                    self.S3_AWS_RESOURCE_BUCKET = config.get(
                        "AWS_S3", "S3_AWS_RESOURCE_BUCKET"
                    )

            if config.has_section("WEBHOOK"):
                if config.has_option("WEBHOOK", "WEBHOOK_RETRIES"):
                    self.WEBHOOK_RETRIES = config.get(
                        "WEBHOOK", "WEBHOOK_RETRIES"
                    )
                if config.has_option("WEBHOOK", "WEBHOOK_SLEEP"):
                    self.WEBHOOK_SLEEP = config.get("WEBHOOK", "WEBHOOK_SLEEP")

        def print_warning(cfg_section, cfg_key, file_val=None, env_val=None):
            if env_val is None:
                env_val = os.environ[cfg_key]
            if config.has_option(cfg_section, cfg_key):
                if file_val is None:
                    file_val = config.get(cfg_section, cfg_key)
                print(
                    "Config for %s from config file with value '%s' will be"
                    " overwritten by environment variable with value '%s'."
                    % (cfg_key, file_val, env_val)
                )

        # Overwrite values with environment variables if exist:
        if os.environ.get("REDIS_SERVER_URL"):
            print_warning("REDIS", "REDIS_SERVER_URL")
            self.REDIS_SERVER_URL = os.environ["REDIS_SERVER_URL"]

        if os.environ.get("REDIS_SERVER_PORT"):
            print_warning("REDIS", "REDIS_SERVER_PORT")
            self.REDIS_SERVER_PORT = os.environ["REDIS_SERVER_PORT"]

        if os.environ.get("REDIS_SERVER_PW"):
            print_warning("REDIS", "REDIS_SERVER_PW", "XXX", "XXX")
            self.REDIS_SERVER_PW = os.environ["REDIS_SERVER_PW"]

        if os.environ.get("REDIS_QUEUE_SERVER_URL"):
            print_warning("QUEUE", "REDIS_QUEUE_SERVER_URL")
            self.REDIS_QUEUE_SERVER_URL = os.environ["REDIS_QUEUE_SERVER_URL"]

        if os.environ.get("REDIS_QUEUE_SERVER_PORT"):
            print_warning("QUEUE", "REDIS_QUEUE_SERVER_PORT")
            self.REDIS_QUEUE_SERVER_PORT = os.environ[
                "REDIS_QUEUE_SERVER_PORT"
            ]

        if os.environ.get("REDIS_QUEUE_SERVER_PW"):
            print_warning("QUEUE", "REDIS_QUEUE_SERVER_PASSWORD", "XXX", "XXX")
            self.REDIS_QUEUE_SERVER_PASSWORD = os.environ[
                "REDIS_QUEUE_SERVER_PW"
            ]


global_config = Configuration()



# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2016-2022 Sören Gebbert and mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""
Resource storage management

TODO: Tests required
"""

import os
from actinia_core.processing.actinia_processing.ephemeral.persistent_processing import (
    PersistentProcessing,
)
from actinia_core.core.common.process_object import Process
from actinia_core.core.common.exceptions import AsyncProcessError


__license__ = "GPLv3"
__author__ = "Sören Gebbert, Anika Weinmann"
__copyright__ = (
    "Copyright 2016-2022, Sören Gebbert and mundialis GmbH & Co. KG"
)
__maintainer__ = "mundialis"


class ResourceStorageDelete(PersistentProcessing):
    """Delete the user specific resource directory"""

    def __init__(self, *args):
        rdc = args[0]
        PersistentProcessing.__init__(self, rdc)
        self.user_resource_storage_path = os.path.join(
            self.config.GRASS_RESOURCE_DIR, self.user_id
        )
        self.olderthan = args[1]

    def _execute(self):
        self._setup()

        if os.path.exists(self.user_resource_storage_path) and os.path.isdir(
            self.user_resource_storage_path
        ):
            if self.olderthan is None:
                # delete all user resources
                executable = "/bin/rm"
                args = ["-rf", self.user_resource_storage_path]
            else:
                # delete all user resources older than X days
                executable = "/usr/bin/find"
                args = [
                    self.user_resource_storage_path,
                    "-mindepth",
                    "1",
                    "-mtime",
                    f"+{self.olderthan}",
                    "-delete",
                ]

            self._run_process(
                Process(
                    exec_type="exec",
                    executable=executable,
                    id="delete_user_specific_resource_directory",
                    executable_params=args,
                )
            )

            if not os.path.exists(self.user_resource_storage_path):
                os.mkdir(self.user_resource_storage_path)
            self.finish_message = "Resource storage successfully removed."
        else:
            raise AsyncProcessError(
                "Resource storage directory <%s> does not exist."
                % self.user_resource_storage_path
            )



# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2016-2022 Sören Gebbert and mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""
Location management

TODO: Integrate into the ephemeral process chain approach
"""

import os
import shutil
from actinia_core.processing.actinia_processing.ephemeral.persistent_processing import (
    PersistentProcessing,
)
from actinia_core.core.common.exceptions import AsyncProcessError

__license__ = "GPLv3"
__author__ = "Sören Gebbert, Carmen Tawalika"
__copyright__ = (
    "Copyright 2016-2022, Sören Gebbert and mundialis GmbH & Co. KG"
)
__maintainer__ = "mundialis"


class PersistentLocationCreator(PersistentProcessing):
    """Create a new location based on EPSG code"""

    def __init__(self, *args):
        PersistentProcessing.__init__(self, *args)

    def _execute(self):
        new_location = self.location_name

        self.location_name = self.config.GRASS_DEFAULT_LOCATION

        self._setup()

        epsg_code = self.request_data["epsg"]

        self._create_temp_database()

        pc = {
            "1": {
                "module": "g.proj",
                "inputs": {"epsg": epsg_code, "location": new_location},
                "flags": "t",
            }
        }

        process_list = self._validate_process_chain(
            process_chain=pc, skip_permission_check=True
        )

        self._create_grass_environment(
            grass_data_base=self.temp_grass_data_base, mapset_name="PERMANENT"
        )

        self._execute_process_list(process_list)

        if os.path.isdir(
            os.path.join(self.temp_grass_data_base, new_location)
        ):
            shutil.move(
                os.path.join(self.temp_grass_data_base, new_location),
                self.grass_user_data_base,
            )
        else:
            raise AsyncProcessError(
                "Unable to create location <%s>" % new_location
            )

        self.finish_message = (
            "Location <%s> successfully created" % new_location
        )



# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2016-2022 Sören Gebbert and mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""
Compute areal categorical statistics on a raster map layer based on an input
polygon.
"""

from actinia_core.processing.actinia_processing.ephemeral_processing import (
    EphemeralProcessing,
)

__license__ = "GPLv3"
__author__ = "Sören Gebbert"
__copyright__ = (
    "Copyright 2016-2022, Sören Gebbert and mundialis GmbH & Co. KG"
)
__maintainer__ = "mundialis"


class ProcessValidation(EphemeralProcessing):
    def __init__(self, *args):
        EphemeralProcessing.__init__(self, *args)

    def _execute(self):
        self._setup()

        process_chain = (
            self._create_temporary_grass_environment_and_process_list()
        )

        result = []
        for process in process_chain:
            result.append(str(process))

        self.finish_message = "Validation successful"

        self.module_results = result



# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2016-2022 Sören Gebbert and mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""
Raster layer resources
"""
from flask import jsonify, make_response
from flask_restful_swagger_2 import swagger
import pickle
from actinia_api.swagger2.actinia_core.apidocs import map_layer_management

from actinia_core.core.common.redis_interface import enqueue_job
from actinia_core.rest.base.endpoint_config import (
    check_endpoint,
    endpoint_decorator,
)
from actinia_core.core.request_parser import glist_parser
from actinia_core.rest.base.resource_base import ResourceBase
from actinia_core.processing.common.map_layer_management import (
    list_raster_layers,
    remove_raster_layers,
    rename_raster_layers,
)

__license__ = "GPLv3"
__author__ = "Sören Gebbert"
__copyright__ = (
    "Copyright 2016-2022, Sören Gebbert and mundialis GmbH & Co. KG"
)
__maintainer__ = "mundialis"


class MapsetLayersResource(ResourceBase):
    """Manage layers of a mapset"""

    def __init__(self, layer_type):
        ResourceBase.__init__(self)
        self.layer_type = layer_type

    def _get(self, location_name, mapset_name):
        """Return a collection of all available layers
        in the provided mapset.

        Optionally can g.list parameters be provided::

            http://<url>?pattern="*"

        Args:
            location_name (str): The name of the location
            mapset_name (str): The name of the mapset

        Return:
            flask.Response: HTTP 200 and a list of layers as JSON document in
                            case of success, HTTP 400 otherwise

        Example::

            {"Process result":
              [ "lsat7_2002_10",
                "lsat7_2002_20",
                "lsat7_2002_30",
                "lsat7_2002_40",
                "lsat7_2002_50",
                "lsat7_2002_61",
                "lsat7_2002_62",
                "lsat7_2002_70",
                "lsat7_2002_80"
              ],
              "Status": "success"
            }

        """
        rdc = self.preprocess(
            has_json=False,
            has_xml=False,
            location_name=location_name,
            mapset_name=mapset_name,
        )

        if rdc:
            args = glist_parser.parse_args()
            rdc.set_user_data((args, self.layer_type))
            enqueue_job(
                self.job_timeout,
                list_raster_layers,
                rdc,
                queue_type_overwrite=True,
            )
            http_code, response_model = self.wait_until_finish()
        else:
            http_code, response_model = pickle.loads(self.response_data)

        return make_response(jsonify(response_model), http_code)

    def _delete(self, location_name, mapset_name):
        """Remove a list of layers identified by a pattern

        The g.remove "pattern" parameters must be provided::

            http://<url>?pattern="*"

        Args:
            location_name (str): The name of the location
            mapset_name (str): The name of the mapset

        Return:
            flask.Response: HTTP 200 in case of success, HTTP 400 otherwise

        """
        rdc = self.preprocess(
            has_json=False,
            has_xml=False,
            location_name=location_name,
            mapset_name=mapset_name,
        )

        if rdc:
            args = glist_parser.parse_args()
            rdc.set_user_data((args, self.layer_type))
            enqueue_job(
                self.job_timeout,
                remove_raster_layers,
                rdc,
                queue_type_overwrite=True,
            )
            http_code, response_model = self.wait_until_finish()
        else:
            http_code, response_model = pickle.loads(self.response_data)

        return make_response(jsonify(response_model), http_code)

    def _put(self, location_name, mapset_name):
        """Rename a list of layers

        The old names and new names must be provided as a
        list of tuples:

            [(a, a_new),(b, b_new),(c, c_new), ...]

        Args:
            location_name (str): The name of the location
            mapset_name (str): The name of the mapset

        Return:
            flask.Response: HTTP 200 in case of success, HTTP 400 otherwise

        """
        rdc = self.preprocess(
            has_json=True,
            has_xml=False,
            location_name=location_name,
            mapset_name=mapset_name,
        )

        # Analyse the name list
        if isinstance(self.request_data, list) is False:
            return self.get_error_response(
                message="Wrong format for layer list"
            )

        if len(self.request_data) == 0:
            return self.get_error_response(message="Empty layer list")

        for name_tuple in self.request_data:
            if (
                isinstance(name_tuple, tuple) is False
                and isinstance(name_tuple, list) is False
            ):
                return self.get_error_response(
                    message="List entry is not a tuple or list"
                )

            if len(name_tuple) != 2:
                return self.get_error_response(
                    message="A tuple of layer names must have 2 entries"
                )

        if rdc:
            args = glist_parser.parse_args()
            rdc.set_user_data((args, self.layer_type))
            enqueue_job(self.job_timeout, rename_raster_layers, rdc)
            http_code, response_model = self.wait_until_finish()
        else:
            http_code, response_model = pickle.loads(self.response_data)

        return make_response(jsonify(response_model), http_code)


class RasterLayersResource(MapsetLayersResource):
    """Manage raster layers of a mapset"""

    def __init__(self):
        MapsetLayersResource.__init__(self, layer_type="raster")

    @endpoint_decorator()
    @swagger.doc(check_endpoint("get", map_layer_management.raster_get_doc))
    def get(self, location_name, mapset_name):
        """Get a list of raster map layer names that are located in a specific
        location/mapset
        """
        return self._get(location_name, mapset_name)

    @endpoint_decorator()
    @swagger.doc(check_endpoint("put", map_layer_management.raster_put_doc))
    def put(self, location_name, mapset_name):
        """Rename a single raster map layer or a list of raster map layers that
        are located in a specific location/mapset
        """
        return self._put(location_name, mapset_name)

    @endpoint_decorator()
    @swagger.doc(
        check_endpoint("delete", map_layer_management.raster_delete_doc)
    )
    def delete(self, location_name, mapset_name):
        """Delete a single raster map layer or a list of raster map layer names
        that are located in a specific location/mapset
        """
        return self._delete(location_name, mapset_name)


class VectorLayersResource(MapsetLayersResource):
    """Manage vector layers of a mapset"""

    def __init__(self):
        MapsetLayersResource.__init__(self, layer_type="vector")

    @endpoint_decorator()
    @swagger.doc(check_endpoint("get", map_layer_management.vector_get_doc))
    def get(self, location_name, mapset_name):
        """Get a list of vector map layer names that are located in a specific
        location/mapset
        """
        return self._get(location_name, mapset_name)

    @endpoint_decorator()
    @swagger.doc(check_endpoint("put", map_layer_management.vector_put_doc))
    def put(self, location_name, mapset_name):
        """Rename a single vector map layer or a list of vector map layers that
        are located in a specific location/mapset
        """
        return self._put(location_name, mapset_name)

    @endpoint_decorator()
    @swagger.doc(
        check_endpoint("delete", map_layer_management.vector_delete_doc)
    )
    def delete(self, location_name, mapset_name):
        """Delete a single vector map layer or a list of vector map layer names
        that are located in a specific location/mapset
        """
        return self._delete(location_name, mapset_name)



# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2016-2018 Sören Gebbert and mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""
Tests: Vector Layer test case

TODO: fix below raster --> vector
"""
from flask.json import loads as json_load
import unittest

try:
    from .test_resource_base import ActiniaResourceTestCaseBase, URL_PREFIX
except ModuleNotFoundError:
    from test_resource_base import ActiniaResourceTestCaseBase, URL_PREFIX


__license__ = "GPLv3"
__author__ = "Sören Gebbert, Anika Weinmann, Guido Riembauer"
__copyright__ = (
    "Copyright 2016-2021, Sören Gebbert and mundialis GmbH & Co. KG"
)
__maintainer__ = "mundialis GmbH & Co. KG"


class RasterLayerTestCase(ActiniaResourceTestCaseBase):
    def test_creation_1(self):
        # Remove potentially existing raster layer
        new_mapset = "raster_test_mapset"
        self.create_new_mapset(new_mapset)

        # Create
        region = {"n": 228500, "s": 215000, "e": 645000, "w": 630000}
        parameter = {"npoints": 1, "zmin": 1, "zmax": 1, "seed": 1}
        self.create_vector_layer(
            "nc_spm_08", new_mapset, "test_layer", region, parameter
        )
        # rv = self.server.post(
        #     URL_PREFIX
        #     + "/locations/nc_spm_08/mapsets/%s/vector_layers/test_layer"
        #     % new_mapset,
        #     headers=self.user_auth_header,
        #     data=json_dumps(
        #         {
        #             "region": {
        #                 "n": 228500,
        #                 "s": 215000,
        #                 "e": 645000,
        #                 "w": 630000,
        #             },
        #             "parameter": {
        #                 "npoints": 1,
        #                 "zmin": 1,
        #                 "zmax": 1,
        #                 "seed": 1,
        #             },
        #         }
        #     ),
        #     content_type="application/json",
        # )
        # print(rv.data)
        # self.assertEqual(
        #     rv.status_code,
        #     200,
        #     "HTML status code is wrong %i" % rv.status_code,
        # )
        # self.assertEqual(
        #     rv.mimetype,
        #     "application/json",
        #     "Wrong mimetype %s" % rv.mimetype
        # )

        # Check
        rv = self.server.get(
            URL_PREFIX
            + "/locations/nc_spm_08/mapsets/%s/vector_layers/test_layer"
            % new_mapset,
            headers=self.user_auth_header,
        )
        print(rv.data)
        self.assertEqual(
            rv.status_code,
            200,
            "HTML status code is wrong %i" % rv.status_code,
        )
        self.assertEqual(
            rv.mimetype, "application/json", "Wrong mimetype %s" % rv.mimetype
        )

        areas = json_load(rv.data)["process_results"]["areas"]
        points = json_load(rv.data)["process_results"]["points"]
        centroids = json_load(rv.data)["process_results"]["centroids"]

        self.assertEqual(areas, "0")
        self.assertEqual(points, "1")
        self.assertEqual(centroids, "0")

        # Delete
        rv = self.server.delete(
            URL_PREFIX
            + "/locations/nc_spm_08/mapsets/%s/vector_layers/test_layer"
            % new_mapset,
            headers=self.user_auth_header,
        )
        print(rv.data)
        self.assertEqual(
            rv.status_code,
            200,
            "HTML status code is wrong %i" % rv.status_code,
        )
        self.assertEqual(
            rv.mimetype, "application/json", "Wrong mimetype %s" % rv.mimetype
        )

        # Delete fail
        rv = self.server.delete(
            URL_PREFIX
            + "/locations/nc_spm_08/mapsets/%s/vector_layers/test_layer"
            % new_mapset,
            headers=self.user_auth_header,
        )
        print(rv.data)
        self.assertEqual(
            rv.status_code,
            400,
            "HTML status code is wrong %i" % rv.status_code,
        )
        self.assertEqual(
            rv.mimetype, "application/json", "Wrong mimetype %s" % rv.mimetype
        )

    def test_layer_info(self):
        rv = self.server.get(
            URL_PREFIX
            + "/locations/nc_spm_08/mapsets/PERMANENT/vector_layers/"
            "boundary_county",
            headers=self.user_auth_header,
        )
        print(rv.data)
        self.assertEqual(
            rv.status_code,
            200,
            "HTML status code is wrong %i" % rv.status_code,
        )
        self.assertEqual(
            rv.mimetype, "application/json", "Wrong mimetype %s" % rv.mimetype
        )

        areas = json_load(rv.data)["process_results"]["areas"]
        nodes = json_load(rv.data)["process_results"]["nodes"]
        centroids = json_load(rv.data)["process_results"]["centroids"]

        self.assertEqual(areas, "926")
        self.assertEqual(nodes, "1114")
        self.assertEqual(centroids, "926")

    def test_layer_info_error_1(self):
        # Raster does not exist
        rv = self.server.get(
            URL_PREFIX
            + "/locations/nc_spm_08/mapsets/PERMANENT/vector_layers/"
            "boundary_county_nope",
            headers=self.user_auth_header,
        )
        print(rv.data)
        self.assertEqual(
            rv.status_code,
            400,
            "HTML status code is wrong %i" % rv.status_code,
        )
        self.assertEqual(
            rv.mimetype, "application/json", "Wrong mimetype %s" % rv.mimetype
        )


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2016-2022 Sören Gebbert and mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""This module manages a raster layer resource as geotiff file that
will be created asynchronously.

The region settings of the mapset
or the raster layer region are used for export.
"""

from actinia_core.processing.actinia_processing.ephemeral.ephemeral_processing_with_export import (
    EphemeralProcessingWithExport,
)

__license__ = "GPLv3"
__author__ = "Sören Gebbert"
__copyright__ = (
    "Copyright 2016-2022, Sören Gebbert and mundialis GmbH & Co. KG"
)
__maintainer__ = "mundialis"


class EphemeralRasterLayerExporter(EphemeralProcessingWithExport):
    """Export a raster layer from a specific mapset as geotiff file.

    The region of the raster layer can be used for export. In this case a
    temporary mapset will be created to modify the region settings safely.
    Hence, this works also in write protected mapsets.
    """

    def __init__(self, rdc):
        """Setup the variables of this class

        Args:
            rdc (ResourceDataContainer): The data container that contains all
                                         required variables for processing

        """

        EphemeralProcessingWithExport.__init__(self, rdc)

        self.raster_name = self.map_name
        self.use_raster_region = self.rdc.user_data

    def _execute(self):
        """Overwrite this function in subclasses

        - Create the resource directory
        - Initialize and create the temporal database and mapset
        - Export the raster layer
        - Cleanup

        """
        # Setup the user credentials and logger
        self._setup()

        # Create and check the resource directory
        self.storage_interface.setup()

        # Check if we have access and create the temporary storage
        self.required_mapsets.append(self.mapset_name)
        self._create_temporary_grass_environment(
            source_mapset_name="PERMANENT"
        )

        # COG bug in GDAL, see https://github.com/OSGeo/gdal/issues/2946 will
        # be fixed in GDAL 3.1.4
        # use r.out.gdal -c to avoid the bug
        format = "COG"
        from osgeo import gdal

        if "COG" not in [
            gdal.GetDriver(i).ShortName for i in range(gdal.GetDriverCount())
        ]:
            format = "GTiff"

        export_dict = {
            "name": self.raster_name + "@" + self.mapset_name,
            "export": {"format": format, "type": "raster"},
        }

        self.resource_export_list.append(export_dict)
        self._export_resources(self.use_raster_region)



# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2016-2018 Sören Gebbert and mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""
Response models
"""
import pickle
import time
from datetime import datetime
from flask import jsonify
from flask_restful_swagger_2 import Schema
from copy import deepcopy
from actinia_api import URL_PREFIX

from actinia_core.core.common.process_chain import GrassModule

__license__ = "GPLv3"
__author__ = "Sören Gebbert, Julia Haas, Guido Riembauer"
__copyright__ = (
    "Copyright 2016-2021, Sören Gebbert and mundialis GmbH & Co. KG"
)
__maintainer__ = "mundialis"


class ProgressInfoModel(Schema):
    """This class defines the model for progress information.

    Progress information is generated in case a process chain with at least one
    commands is processed.
    """

    type = "object"
    properties = {
        "step": {
            "type": "integer",
            "format": "int64",
            "description": "The current processing step",
        },
        "num_of_steps": {
            "type": "integer",
            "format": "int64",
            "description": "The total number of processing steps",
        },
        "sub_step": {
            "type": "integer",
            "format": "int64",
            "description": "The current sub step of the current processing "
            "step",
        },
        "num_of_sub_steps": {
            "type": "integer",
            "format": "int32",
            "description": "The total number of sub steps of the current "
            "processing step",
        },
    }
    required = ["step", "num_of_steps"]

    example = {"num_of_steps": 6, "step": 6}


class ProcessLogModel(Schema):
    """This class defines the model for Unix process information.

    Each time a Unix process
    is invoked, this model must be used to inform the user of the return
    process state. Each process has parameters, stdout/stderr output and a
    return value. This model is not designed to inform about running processes,
    but about finished processes. The return_code of 0 indicates that it ran
    successfully. A negative value -N indicates that the child was terminated
    by signal N (POSIX only; see also
    https://en.wikipedia.org/wiki/Signal_(IPC)#Default_action).
    """

    type = "object"
    properties = {
        "id": {"type": "string", "description": "The ID of the executable"},
        "executable": {
            "type": "string",
            "description": "The name of the executable",
        },
        "parameter": {
            "type": "array",
            "items": {"type": "string"},
            "description": "The parameter of the executable",
        },
        "stdout": {
            "type": "string",
            "description": "The stdout output of the executable",
        },
        "stderr": {
            "type": "array",
            "items": {"type": "string"},
            "description": "The stderr output of the executable as list of "
            "strings",
        },
        "return_code": {
            "type": "number",
            "format": "int32",
            "description": "The return code of the executable",
        },
        "run_time": {
            "type": "number",
            "format": "float",
            "description": "The runtime of the executable in seconds",
        },
        "mapset_size": {
            "type": "number",
            "format": "float",
            "description": "The size of the mapset in bytes",
        },
    }
    required = ["executable", "parameter", "stdout", "stderr", "return_code"]

    example = {
        "executable": "g.list",
        "parameter": ["type=raster", "mapset=PERMANENT"],
        "return_code": 0,
        "run_time": 0.05017662048339844,
        "stderr": [""],
        "stdout": "aspect\nbasin_50K\n",
    }


class UrlModel(Schema):
    """URL schema that points to the status URL
    of the resource and to all generated resources (files, images, ...).

    """

    type = "object"
    properties = {
        "status": {
            "type": "string",
            "description": "The URL to check the status of the resource",
        },
        "resources": {
            "type": "array",
            "items": {"type": "string"},
            "description": "A list of URLs to generated resources, that may "
            "be GeoTiff files, vector files, ASCII files or png images",
        },
    }
    required = ["status", "resources"]
    example = {
        "resources": [
            f"http://localhost{URL_PREFIX}/resource/user/resource_id-4846cbcc-"
            "3918-4654-bf4d-7e1ba2b59ce6/my_slope.tiff"
        ],
        "status": f"http://localhost{URL_PREFIX}/resources/user/resource_id-"
        "4846cbcc-3918-4654-bf4d-7e1ba2b59ce6",
    }


class SimpleResponseModel(Schema):
    """
    Response schema that is used in cases that no asynchronous run was
    performed and the generated information is very simple.

    """

    type = "object"
    properties = {
        "status": {
            "type": "string",
            "description": "The status of the resource, values: accepted, "
            "running, finished, terminated, error",
        },
        "message": {
            "type": "string",
            "description": "A simple message to describes the status of the "
            "resource",
        },
    }
    required = ["status", "message"]


class MapsetListResponseModel(Schema):
    """
    Response schema that is used to list all mapsets available to the user.
    """

    type = "object"
    properties = {
        "status": {
            "type": "string",
            "description": "The status of the request",
        },
        "available_mapsets": {
            "type": "array",
            "items": {"type": "string"},
            "description": (
                "The names of all available mapsets with"
                " corresponding locations in the global database"
            ),
        },
    }
    required = ["status", "available_mapsets"]
    example = {
        "status": "success",
        "available_mapsets": [
            "nc_spm_08/PERMANENT",
            "latlong_wgs84/PERMANENT",
        ],
    }


class LockedMapsetListResponseModel(Schema):
    """Response schema that is used to list all locked mapsets."""

    type = "object"
    properties = {
        "status": {
            "type": "string",
            "description": "The status of the request",
        },
        "locked_mapsets_list": {
            "type": "array",
            "items": {"type": "string"},
            "description": "The names of all locked mapsets",
        },
        "message": {
            "type": "string",
            "description": "A simple message to describes the status of the "
            "resource",
        },
    }
    required = ["status", "locked_mapsets_list", "message"]
    example = {
        "status": "success",
        "locked_mapsets_list": ["utm32n/test_mapset"],
        "message": "number of locked mapsets: 1",
    }


class ApiInfoModel(Schema):
    """Response schema that contains API information of the called endpoint.

    This information is used in the ProcessResponseModel schema and
    important for the accounting system.

    """

    type = "object"
    properties = {
        "endpoint": {
            "type": "string",
            "description": "The endpoint of the API call",
        },
        "method": {
            "type": "string",
            "description": "The HTTP method of the request",
        },
        "path": {
            "type": "string",
            "description": "The path of the REST API call",
        },
        "request_url": {"type": "string", "description": "The request URL"},
        "post_url": {"type": "string", "description": "The post URL"},
    }
    required = ["endpoint", "method", "path", "request_url"]

    example = {
        "endpoint": "asyncephemeralresource",
        "method": "POST",
        "path": f"{URL_PREFIX}/locations/nc_spm_08/processing_async",
        "request_url": f"http://localhost{URL_PREFIX}/locations/nc_spm_08/"
        "processing_async",
    }


class ExceptionTracebackModel(Schema):
    """
    Response schema that contains python3 exception information of the called
    endpoint in case an exception was raised.

    This information is required to debug the REST API.
    """

    type = "object"
    properties = {
        "message": {
            "type": "string",
            "description": "The message that was send with the exception",
        },
        "type": {"type": "string", "description": "The type of the exception"},
        "traceback": {
            "type": "array",
            "items": {"type": "string"},
            "description": "The full traceback of the exception",
        },
    }
    required = ["message", "type", "traceback"]

    example = {
        "message": "Error",
        "type": "exceptions.Exception",
        "traceback": 'File "main.py", line 2, in <module>\n    '
        'raise Exception("Error")\n',
    }


class ProcessingResponseModel(Schema):
    """This is the base class for ALL response models.

    This class or its derivatives must be used in all responses that run
    GRASS modules or Unix programs. Derivatives only differ in the
    *process_result* schema.
    """

    type = "object"
    properties = {
        "status": {
            "type": "string",
            "description": "The status of the response",
        },
        "user_id": {
            "type": "string",
            "description": "The id of the user that issued a request",
        },
        "resource_id": {
            "type": "string",
            "description": "The unique resource id",
        },
        "queue": {
            "type": "string",
            "description": "The name of the queue in which the job is queued",
        },
        "process_log": {
            "type": "array",
            "items": ProcessLogModel,
            "description": "A list of ProcessLogModels",
        },
        "process_chain_list": {
            "type": "array",
            "items": GrassModule,
            "description": "The list of GRASS modules that were used in the "
            "processing",
        },
        "process_results": {
            "type": "string",
            "description": "An arbitrary class that stores the processing "
            "results",
        },
        "progress": ProgressInfoModel,
        "message": {
            "type": "string",
            "description": "Message for the user, maybe status, finished or "
            "error message",
        },
        "exception": ExceptionTracebackModel,
        "accept_timestamp": {
            "type": "number",
            "format": "double",
            "description": "The acceptance timestamp in seconds of the "
            "response",
        },
        "accept_datetime": {
            "type": "string",
            "description": "The acceptance timestamp of the response in human "
            "readable format",
        },
        "timestamp": {
            "type": "number",
            "format": "double",
            "description": "The current timestamp in seconds of the response",
        },
        "time_delta": {
            "type": "number",
            "format": "double",
            "description": "The time delta of the processing in seconds",
        },
        "datetime": {
            "type": "string",
            "description": "The current timestamp of the response in human "
            "readable format",
        },
        "http_code": {
            "type": "number",
            "format": "int32",
            "description": "The HTTP code of the response",
        },
        "urls": UrlModel,
        "api_info": ApiInfoModel,
    }
    required = [
        "status",
        "user_id",
        "resource_id",
        "timestamp",
        "datetime",
        "accept_timestamp",
        "accept_datetime",
        "message",
    ]

    example = {
        "accept_datetime": "2017-05-24 22:37:21.607255",
        "accept_timestamp": 1495658241.607252,
        "api_info": {
            "endpoint": "asyncephemeralresource",
            "method": "POST",
            "path": "/locations/nc_spm_08/processing_async",
            "request_url": f"http://localhost{URL_PREFIX}/locations/nc_spm_08/"
            "processing_async",
        },
        "datetime": "2017-05-24 22:37:21.608717",
        "http_code": 200,
        "message": "Resource accepted",
        "process_results": {},
        "resource_id": "resource_id-2be8cafe-b451-46a0-be15-f61d95c5efa1",
        "status": "accepted",
        "timestamp": 1495658241.608716,
        "urls": {
            "resources": [],
            "status": f"http://localhost{URL_PREFIX}/resources/admin/"
            "resource_id-2be8cafe-b451-46a0-be15-f61d95c5efa1",
        },
        "user_id": "admin",
    }


class ProcessingErrorResponseModel(ProcessingResponseModel):
    """
    Response schema that returns error messages from a failed process execution
    that includes debug information (python traceback) and process logs.
    """

    type = "object"
    properties = deepcopy(ProcessingResponseModel.properties)
    required = deepcopy(ProcessingResponseModel.required)
    example = {
        "accept_datetime": "2018-05-06 22:02:14.323815",
        "accept_timestamp": 1525636934.3238132,
        "api_info": {
            "endpoint": "mapsetmanagementresourceuser",
            "method": "GET",
            "path": "/locations/nc_spm_08/mapsets/PERMANE/info",
            "request_url": f"http://localhost{URL_PREFIX}/locations/nc_spm_08/"
            "mapsets/PERMANENT/info",
        },
        "datetime": "2018-05-06 22:02:14.398927",
        "exception": {
            "message": "AsyncProcessError:  Error while running executable "
            "<g.region>",
            "traceback": [
                '  File "/home/soeren/src/GRaaS/actinia_venv/lib/python3.5/'
                "site-packages/actinia_core-0.0.post0.dev37+g216eeae.dirty-"
                'py3.5.egg/actinia_core/resources/ephemeral_processing.py", '
                "line 1063, in _run_executable\n    raise AsyncProcessError("
                '"Error while running executable <%s>" % process.executable)\n'
            ],
            "type": "<class 'actinia_core.core.common.exceptions."
            "AsyncProcessError'>",
        },
        "http_code": 400,
        "message": "AsyncProcessError:  Error while running executable "
        "<g.region>",
        "process_chain_list": [
            {
                "1": {"flags": "ug3", "module": "g.region"},
                "2": {"flags": "fw", "module": "g.proj"},
            }
        ],
        "process_log": [
            {
                "executable": "g.region",
                "parameter": ["-ug3"],
                "return_code": 1,
                "run_time": 0.05020904541015625,
                "stderr": [
                    "ERROR: MAPSET PERMANE not found at /home/soeren/actinia/"
                    "workspace/temp_db/gisdbase_5c4c13bce6e54207aea2e1705cba"
                    "0b8b/nc_spm_08",
                    "",
                ],
                "stdout": "",
            }
        ],
        "progress": {"num_of_steps": 2, "step": 1},
        "resource_id": "resource_id-79608249-521c-4a98-9e1f-9201f693870b",
        "status": "error",
        "time_delta": 0.07516026496887207,
        "timestamp": 1525636934.3989098,
        "urls": {
            "resources": [],
            "status": f"http://localhost{URL_PREFIX}/resources/user/"
            "resource_id-79608249-521c-4a98-9e1f-9201f693870b",
        },
        "user_id": "user",
    }


class ProcessingResponseListModel(Schema):
    """Response schema that represent a list of ProcessingResponseModel's"""

    type = "object"
    properties = {
        "resource_list": {
            "type": "array",
            "items": ProcessingResponseModel,
            "description": "A list of ProcessingResponseModel objects",
        }
    }
    required = ["resource_list"]


class StorageModel(Schema):
    """This class defines the model to inform about available storage
    that is used for caching or user specific resource storage.

    It is used as schema to define the *process_result* in a
    ProcessingResponseModel derivative.
    """

    type = "object"
    properties = {
        "used": {
            "type": "number",
            "format": "int64",
            "description": "The used storage in bytes",
        },
        "free": {
            "type": "number",
            "format": "int64",
            "description": "The free storage in bytes",
        },
        "quota": {
            "type": "number",
            "format": "int64",
            "description": "The current quota in bytes",
        },
        "free_percent": {
            "type": "number",
            "format": "int32",
            "description": "The free storage in percent",
        },
    }
    required = ["used", "free", "quota", "free_percent"]


class UnivarResultModel(Schema):
    """
    Response schema for the result of univariate computations of raster layers.

    It is used as schema to define the *process_result* in a
    ProcessingResponseModel derivative.
    """

    type = "object"
    properties = {
        "name": {
            "type": "string",
            "description": "The name of the raster resource",
        },
        "cells": {
            "type": "number",
            "format": "double",
        },
        "coeff_var": {
            "type": "number",
            "format": "double",
        },
        "max": {
            "type": "number",
            "format": "double",
        },
        "mean": {
            "type": "number",
            "format": "double",
        },
        "mean_of_abs": {
            "type": "number",
            "format": "double",
        },
        "min": {
            "type": "number",
            "format": "double",
        },
        "n": {
            "type": "number",
            "format": "double",
        },
        "null_cells": {
            "type": "number",
            "format": "double",
        },
        "range": {
            "type": "number",
            "format": "double",
        },
        "stddev": {
            "type": "number",
            "format": "double",
        },
        "sum": {
            "type": "number",
            "format": "double",
        },
        "variance": {
            "type": "number",
            "format": "double",
        },
    }
    # If a map is empty, r.univar will return nothing, hence no required
    # variables
    # required = ['name', 'cells', 'coeff_var', 'max', 'mean', 'mean_of_abs',
    #            'min','n', 'null_cells', 'range', 'stddev', 'sum', 'variance']


class CategoricalStatisticsResultModel(Schema):
    """Response schema for the result of r.stats computations of raster layers.

    It is used as schema to define the *process_result* in a
    ProcessingResponseModel derivative.
    """

    type = "object"
    properties = {
        "cat": {"type": "string", "description": "The raster category"},
        "name": {
            "type": "string",
            "description": "The name of raster category",
        },
        "area": {
            "type": "number",
            "format": "double",
            "description": "The size of the area in square meters",
        },
        "cell_count": {
            "type": "number",
            "format": "double",
            "description": "The number of cells that have the raster category",
        },
        "percent": {
            "type": "number",
            "format": "double",
            "description": "The percentage of the area",
        },
    }
    required = ["cat", "name", "area", "cell_count", "percent"]


class RegionModel(Schema):
    """Output of GRASS GIS module g.region -gu3

    It is used as schema to define the *process_result* in a
    ProcessingResponseModel derivative.

    GRASS 7.8.dev (latlong_wgs84):~ > g.region -gu3
    projection=3
    zone=0
    n=80
    s=0
    w=0
    e=120
    t=50
    b=0
    nsres=10
    nsres3=10
    ewres=10
    ewres3=10
    tbres=10
    rows=8
    rows3=8
    cols=12
    cols3=12
    depths=5
    cells=96
    cells3=480

    """

    type = "object"
    properties = {
        "projection": {
            "type": "number",
            "format": "integer",
        },
        "zone": {
            "type": "number",
            "format": "integer",
        },
        "n": {
            "type": "number",
            "format": "double",
        },
        "s": {
            "type": "number",
            "format": "double",
        },
        "w": {
            "type": "number",
            "format": "double",
        },
        "e": {
            "type": "number",
            "format": "double",
        },
        "t": {
            "type": "number",
            "format": "double",
        },
        "b": {
            "type": "number",
            "format": "double",
        },
        "nsres": {
            "type": "number",
            "format": "double",
        },
        "nsres3": {
            "type": "number",
            "format": "double",
        },
        "ewres": {
            "type": "number",
            "format": "double",
        },
        "ewres3": {
            "type": "number",
            "format": "double",
        },
        "tbres": {
            "type": "number",
            "format": "double",
        },
        "rows": {
            "type": "number",
            "format": "integer",
        },
        "rows3": {
            "type": "number",
            "format": "integer",
        },
        "cols": {
            "type": "number",
            "format": "integer",
        },
        "cols3": {
            "type": "number",
            "format": "integer",
        },
        "depths": {
            "type": "number",
            "format": "integer",
        },
        "cells": {
            "type": "number",
            "format": "integer",
        },
        "cells3": {
            "type": "number",
            "format": "integer",
        },
    }

    example = {
        "b": 0.0,
        "cells": 29535,
        "cells3": 29535,
        "cols": 179,
        "cols3": 179,
        "depths": 1,
        "e": 639530.0,
        "ewres": 10.0,
        "ewres3": 10.0,
        "n": 221230.0,
        "nsres": 10.0,
        "nsres3": 10.0,
        "projection": 99,
        "rows": 165,
        "rows3": 165,
        "s": 219580.0,
        "t": 1.0,
        "tbres": 1.0,
        "w": 637740.0,
        "zone": 0,
    }


class MapsetInfoModel(Schema):
    """Schema for projection and region information from a specific mapset.

    It is used as schema to define the *process_result* in a
    ProcessingResponseModel derivative.
    """

    type = "object"
    properties = {
        "projection": {
            "type": "string",
            "description": "The location projection WKT string",
        },
        "region": RegionModel,
    }
    required = ["projection", "region"]

    example = {
        "projection": 'PROJCS["NAD83(HARN) / North Carolina",GEOGCS['
        '"NAD83(HARN)",DATUM["NAD83_High_Accuracy_Reference_Network",'
        'SPHEROID["GRS1980",6378137,298.257222101,AUTHORITY["EPSG","7019"]],'
        'TOWGS84[0,0,0,0,0,0,0],AUTHORITY["EPSG","6152"]],PRIMEM['
        '"Greenwich",0,AUTHORITY["EPSG","8901"]],UNIT["degree",'
        '0.0174532925199433,AUTHORITY["EPSG","9122"]],AUTHORITY['
        '"EPSG","4152"]],PROJECTION["Lambert_Conformal_Conic_2SP"],'
        'PARAMETER["standard_parallel_1",36.16666666666666],'
        'PARAMETER["standard_parallel_2",34.33333333333334],'
        'PARAMETER["latitude_of_origin",33.75],PARAMETER['
        '"central_meridian",-79],PARAMETER["false_easting",'
        '609601.22],PARAMETER["false_northing",0],UNIT["metre",1,'
        'AUTHORITY["EPSG","9001"]],AXIS["X",EAST],'
        'AXIS["Y",NORTH],AUTHORITY["EPSG","3358"]]\n',
        "region": {
            "b": 0.0,
            "cells": 29535,
            "cells3": 29535,
            "cols": 179,
            "cols3": 179,
            "depths": 1,
            "e": 639530.0,
            "ewres": 10.0,
            "ewres3": 10.0,
            "n": 221230.0,
            "nsres": 10.0,
            "nsres3": 10.0,
            "projection": 99,
            "rows": 165,
            "rows3": 165,
            "s": 219580.0,
            "t": 1.0,
            "tbres": 1.0,
            "w": 637740.0,
            "zone": 0,
        },
    }


class MapsetInfoResponseModel(ProcessingResponseModel):
    """Response schema that includes projection and region information
    about a specific mapset as processing results.

    This schema is a derivative of the ProcessingResponseModel that defines a
    different *process_results* schema.
    """

    type = "object"
    properties = deepcopy(ProcessingResponseModel.properties)
    properties["process_results"] = MapsetInfoModel
    required = deepcopy(ProcessingResponseModel.required)
    example = {
        "accept_datetime": "2018-05-02 10:53:20.254387",
        "accept_timestamp": 1525258400.2543857,
        "api_info": {
            "endpoint": "mapsetmanagementresourceuser",
            "method": "GET",
            "path": f"{URL_PREFIX}/locations/ECAD/mapsets/PERMANENT/info",
            "request_url": f"http://localhost{URL_PREFIX}/locations/ECAD/"
            "mapsets/PERMANENT/info",
        },
        "datetime": "2018-05-02 10:53:20.392509",
        "http_code": 200,
        "message": "Processing successfully finished",
        "process_chain_list": [
            {
                "1": {"flags": "ug3", "module": "g.region"},
                "2": {"flags": "fw", "module": "g.proj"},
            }
        ],
        "process_log": [
            {
                "executable": "g.region",
                "parameter": ["-ug3"],
                "return_code": 0,
                "run_time": 0.05018115043640137,
                "stderr": [""],
                "stdout": "...",
            },
            {
                "executable": "g.proj",
                "parameter": ["-fw"],
                "return_code": 0,
                "run_time": 0.05016374588012695,
                "stderr": [""],
                "stdout": "...",
            },
        ],
        "process_results": {
            "projection": 'GEOGCS["WGS 84",DATUM["WGS_1984",SPHEROID["WGS 84",'
            '6378137,298.257223563,AUTHORITY["EPSG","7030"]],'
            'AUTHORITY["EPSG","6326"]],PRIMEM["Greenwich",0,'
            'AUTHORITY["EPSG","8901"]],UNIT["degree",'
            '0.0174532925199433,AUTHORITY["EPSG","9122"]],'
            'AUTHORITY["EPSG","4326"]]\n',
            "region": {
                "b": 0.0,
                "cells": 1,
                "cells3": 1,
                "cols": 1,
                "cols3": 1,
                "depths": 1,
                "e": 1.0,
                "ewres": 1.0,
                "ewres3": 1.0,
                "n": 1.0,
                "nsres": 1.0,
                "nsres3": 1.0,
                "projection": 3,
                "rows": 1,
                "rows3": 1,
                "s": 0.0,
                "t": 1.0,
                "tbres": 1.0,
                "w": 0.0,
                "zone": 0,
            },
        },
        "progress": {"num_of_steps": 2, "step": 2},
        "resource_id": "resource_id-2222cdb7-06f5-460d-a38f-5745a3c3b518",
        "status": "finished",
        "time_delta": 0.13819408416748047,
        "timestamp": 1525258400.392495,
        "urls": {
            "resources": [],
            "status": f"http://localhost{URL_PREFIX}/resources/user/"
            "resource_id-2222cdb7-06f5-460d-a38f-5745a3c3b518",
        },
        "user_id": "user",
    }


class RasterAreaStatsResponseModel(ProcessingResponseModel):
    """Response schema for a list of categorical statistics.

    This schema is a derivative of the ProcessingResponseModel that defines a
    different *process_results* schema.
    """

    type = "object"
    properties = deepcopy(ProcessingResponseModel.properties)
    properties["process_results"] = {}
    properties["process_results"]["type"] = "array"
    properties["process_results"]["items"] = CategoricalStatisticsResultModel
    required = deepcopy(ProcessingResponseModel.required)
    example = {}


class AreaUnivarResultModel(Schema):
    """Response schema for the result of univariate computations of raster
    layers based on a vector area.

    It is used as schema to define the process_result in a
    ProcessingResponseModel.

    cat
    fid
    raster_number
    raster_minimum
    raster_maximum
    raster_range
    raster_average
    raster_median
    raster_stddev
    raster_sum
    raster_variance
    raster_coeff_var

    """

    type = "object"
    properties = {
        "fid": {
            "type": "string",
            "description": "Field id from the polygon of the vector map layer "
            "used for area stats computation",
        },
        "cat": {
            "type": "string",
            "description": "The category id from the polygon of the vector map"
            " layer used for area stats computation",
        },
        "raster_number": {
            "type": "number",
            "format": "double",
        },
        "raster_minimum": {
            "type": "number",
            "format": "double",
        },
        "raster_maximum": {
            "type": "number",
            "format": "double",
        },
        "raster_range": {
            "type": "number",
            "format": "double",
        },
        "raster_average": {
            "type": "number",
            "format": "double",
        },
        "raster_median": {
            "type": "number",
            "format": "double",
        },
        "raster_stddev": {
            "type": "number",
            "format": "double",
        },
        "raster_sum": {
            "type": "number",
            "format": "double",
        },
        "raster_variance": {
            "type": "number",
            "format": "double",
        },
        "raster_coeff_var": {
            "type": "number",
            "format": "double",
        },
    }


class RasterAreaUnivarStatsResponseModel(ProcessingResponseModel):
    """Response schema for resources that generate area univariate result lists
     as processing results.

    This schema is a derivative of the ProcessingResponseModel that defines a
    different *process_results* schema.
    """

    type = "object"
    properties = deepcopy(ProcessingResponseModel.properties)
    properties["process_results"] = {}
    properties["process_results"]["type"] = "array"
    properties["process_results"]["items"] = AreaUnivarResultModel
    required = deepcopy(ProcessingResponseModel.required)
    # required.append("process_results")
    example = {}


class StorageResponseModel(ProcessingResponseModel):
    """Response schema for storage management calls.

    This schema is a derivative of the ProcessingResponseModel that defines a
    different *process_results* schema.
    """

    type = "object"
    properties = deepcopy(ProcessingResponseModel.properties)
    properties["process_results"] = StorageModel
    required = deepcopy(ProcessingResponseModel.required)
    # required.append("process_results")
    example = {
        "accept_datetime": "2018-05-02 10:51:27.988969",
        "accept_timestamp": 1525258287.9889677,
        "api_info": {
            "endpoint": "syncresourcestorageresource",
            "method": "GET",
            "path": "/resource_storage",
            "request_url": "http://localhost:8080/resource_storage",
        },
        "datetime": "2018-05-02 10:51:28.100659",
        "http_code": 200,
        "message": "Resource storage size successfully computed",
        "process_chain_list": [],
        "process_log": [
            {
                "executable": "/usr/bin/du",
                "parameter": ["-sb", "/actinia/resources/superadmin"],
                "return_code": 0,
                "run_time": 0.05042886734008789,
                "stderr": [""],
                "stdout": "4096\t/actinia/resources/superadmin\n",
            }
        ],
        "process_results": {
            "free": 107374178304,
            "free_percent": 99,
            "quota": 107374182400,
            "used": 4096,
        },
        "progress": {"num_of_steps": 0, "step": 1},
        "resource_id": "resource_id-f9463d91-04a6-497c-b107-37c4ee013530",
        "status": "finished",
        "time_delta": 0.11174893379211426,
        "timestamp": 1525258288.1006439,
        "urls": {
            "resources": [],
            "status": "http://localhost:8080/resources/superadmin/resource_id-"
            "f9463d91-04a6-497c-b107-37c4ee013530",
        },
        "user_id": "superadmin",
    }


class StringListProcessingResultResponseModel(ProcessingResponseModel):
    """Response schema for resources that generate string list
     as processing results.

    This schema is a derivative of the ProcessingResponseModel that defines a
    different *process_results* schema.
    """

    type = "object"
    properties = deepcopy(ProcessingResponseModel.properties)
    properties["process_results"] = {}
    properties["process_results"]["type"] = "array"
    properties["process_results"]["items"] = {"type": "string"}
    required = deepcopy(ProcessingResponseModel.required)
    # required.append("process_results")
    example = {
        "accept_datetime": "2018-05-02 12:02:20.746845",
        "accept_timestamp": 1525255340.746844,
        "api_info": {
            "endpoint": "listmapsetsresource",
            "method": "GET",
            "path": "/locations/nc_spm_08/mapsets",
            "request_url": "http://localhost:5000/locations/nc_spm_08/mapsets",
        },
        "datetime": "2018-05-02 12:02:20.861017",
        "http_code": 200,
        "message": "Processing successfully finished",
        "process_chain_list": [
            {
                "1": {
                    "flags": "l",
                    "inputs": {"separator": "newline"},
                    "module": "g.mapsets",
                }
            }
        ],
        "process_log": [
            {
                "executable": "g.mapsets",
                "parameter": ["separator=newline", "-l"],
                "return_code": 0,
                "run_time": 0.0501406192779541,
                "stderr": ["Available mapsets:", ""],
                "stdout": "PERMANENT\nSource_A\nSource_B\nTarget\nlandsat\n"
                "raster_test_mapset\nuser1\n",
            }
        ],
        "process_results": [
            "PERMANENT",
            "Source_A",
            "Source_B",
            "Target",
            "landsat",
            "raster_test_mapset",
            "user1",
        ],
        "progress": {"num_of_steps": 1, "step": 1},
        "resource_id": "resource_id-715530fe-53e2-4b2b-87b2-3777c90fec7c",
        "status": "finished",
        "time_delta": 0.11421895027160645,
        "timestamp": 1525255340.8610027,
        "urls": {
            "resources": [],
            "status": "http://localhost:5000/resources/user/resource_id-"
            "715530fe-53e2-4b2b-87b2-3777c90fec7c",
        },
        "user_id": "user",
    }


def create_response_from_model(
    response_model_class=ProcessingResponseModel,
    status=None,
    user_id=None,
    resource_id=None,
    queue=None,
    iteration=None,
    process_log=None,
    progress=None,
    results=None,
    message=None,
    http_code=None,
    status_url=None,
    orig_time=None,
    orig_datetime=None,
    resource_urls=[],
    api_info=None,
    process_chain_list=[],
    exception=None,
    resp_type="pickle",
):
    """Create a dictionary and its pickle or JSON representation to represent
    response information

    This function is used to create almost all responses of Actinia Core

    Args:
        response_model_class (class): The response model class that must be
                                      applied. By default this is the
                                      ProcessingResponseModel class, that is
                                      the base for all responses that involve
                                      executables in Actinia Core. All classes
                                      that are accepted must be derived from
                                      ProcessingResponseModel and specify the
                                      "process_results" field. By default a
                                      string is required as data type for
                                      "process_results".
        status (str): One of: accepted, running, finished, error
        user_id (str): The user id
        resource_id (str): The resource id
        queue (str): The name of the queue
        iteration (int): The iteration of the job
        process_log (dict, str, list): The log from the running GRASS module
        progress (ProgressInfoModel): Progress information
        results (dict): The results of processing steps as Python data types
        message (str): The message from the running GRASS module or an error
                       message
        http_code (int): The HTTP status code
        status_url (str): The url of this request
        orig_time (time): The time of origin (seconds)
        orig_datetime (datetime): The datetime of origin (datetime format)
        resource_urls ([str]): The list of url of the new created resources
        api_info (ApiInfoModel): Information about the API call, important for
                                 accounting
        process_chain_list ([ProcessChainModel]): The list of process chains
        resp_type (str): What type of response, "pickle" or "json"

    Returns:
        A pickle string or json string

    """
    # if issubclass(response_model_class, ProcessingResponseModel) is False:
    #    raise IOError

    # if webhook auth is given set pw to XXX
    for i in range(len(process_chain_list)):
        if (
            "webhooks" in process_chain_list[i]
            and "auth" in process_chain_list[i]["webhooks"]
        ):
            webhook_auth = process_chain_list[i]["webhooks"]["auth"]
            process_chain_list[i]["webhooks"][
                "auth"
            ] = f"{webhook_auth.split(':')[0]}:XXX"

    resp_dict = response_model_class(
        status=status,
        user_id=user_id,
        resource_id=resource_id,
        queue=str(queue),
        # iteration=iteration,
        accept_timestamp=orig_time,
        accept_datetime=orig_datetime,
        timestamp=time.time(),
        datetime=str(datetime.now()),
        message=message,
        time_delta=time.time() - orig_time,
    )

    if process_log is not None:
        resp_dict["process_log"] = process_log
    if progress is not None:
        resp_dict["progress"] = progress
    if results is not None:
        resp_dict["process_results"] = results
    if process_chain_list is not None:
        resp_dict["process_chain_list"] = process_chain_list
    if http_code is not None:
        resp_dict["http_code"] = http_code
    if exception is not None:
        resp_dict["exception"] = exception
    if http_code is not None:
        resp_dict["urls"] = UrlModel(
            resources=resource_urls, status=str(status_url)
        )
    if api_info is not None:
        resp_dict["api_info"] = api_info
    if iteration is not None:
        resp_dict["iteration"] = iteration

    if resp_type == "pickle":
        return pickle.dumps([http_code, resp_dict])
    else:
        return jsonify(resp_dict)


class UserListResponseModel(Schema):
    """Response schema that is used to list all users."""

    type = "object"
    properties = {
        "status": {
            "type": "string",
            "description": "The status of the request",
        },
        "user_list": {
            "type": "array",
            "items": {"type": "string"},
            "description": "The names of all users",
        },
    }
    required = ["status", "user_list"]
    example = {"status": "success", "user_list": ["actinia-gdi"]}


class UserInfoResponseModel(Schema):
    """Response schema that is used to show information about a user."""

    type = "object"
    properties = {
        "status": {
            "type": "string",
            "description": "The status of the request",
        },
        "permissions": {
            "type": "object",
            "properties": {
                "cell_limit": {
                    "type": "string",
                    "description": "The limit of number of raster cells the "
                    "user is allowed to process",
                },
                "process_num_limit": {
                    "type": "string",
                    "description": "The limit of number of processes the user "
                    "is allowed to integrate into one process chain",
                },
                "process_time_limit": {
                    "type": "string",
                    "description": "The time a process must not exceed",
                },
                "accessible_datasets": {
                    "type": "object",
                    "properties": {},
                    "description": "The persistent GRASS GIS databases the "
                    "user is allowed to use. Contains one object for each "
                    "location name with an array of strings containing "
                    "all allowed mapset names. See example for more "
                    "information.",
                },
                "accessible_modules": {
                    "type": "array",
                    "items": {"type": "string"},
                    "description": "The GRASS GIS modules the user is allowed"
                    " to use",
                },
            },
            "description": "The names of all users",
        },
        "user_id": {
            "type": "string",
            "description": "The identifier of the user",
        },
        "user_role": {"type": "string", "description": "The role of the user"},
        "user_group": {
            "type": "string",
            "description": "The group of the user",
        },
    }
    required = ["status"]
    example = {
        "Permissions": {
            "accessible_datasets": {
                "nc_spm_08": ["PERMANENT", "user1", "landsat"]
            },
            "accessible_modules": [
                "r.blend",
                "r.buffer",
                "...",
                "t.rast.univar",
                "importer",
                "exporter",
            ],
            "cell_limit": 100000000000,
            "process_num_limit": 1000,
            "process_time_limit": 31536000,
        },
        "Status": "success",
        "User group": "superadmin",
        "User id": "actinia-gdi",
        "User role": "superadmin",
    }


class LinkResponseModel(Schema):
    """Response schema that is used in cases that no asynchronous run was
    performed and the generated information is very simple.

    """

    type = "object"
    properties = {
        "status": {
            "type": "string",
            "description": "The status of the resource, values: accepted, "
            "running, finished, terminated, error",
        },
        "message": {
            "type": "string",
            "description": "A simple message to describes the status of the "
            "resource",
        },
        "links": {"type": "list", "description": "A list of related links"},
    }
    required = ["status", "message"]



# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2016-2019 Sören Gebbert and mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""
Base class for GRASS GIS REST API tests
"""
import base64
import os
import requests
import time
import unittest
from flask.json import loads as json_loads
from flask.json import dumps as json_dumps
from werkzeug.datastructures import Headers
from actinia_api import URL_PREFIX

from .health_check import health_check
from .version import version
from actinia_core.core.common.app import flask_app
from actinia_core.core.common import redis_interface
from actinia_core.core.common.config import global_config
from actinia_core.core.common.user import ActiniaUser
from actinia_core.core.common.process_queue import create_process_queue

__license__ = "GPLv3"
__author__ = "Sören Gebbert"
__copyright__ = (
    "Copyright 2016-2018, Sören Gebbert and mundialis GmbH & Co. KG"
)
__maintainer__ = "mundialis"


class ActiniaRequests(object):
    """Requests to a actinia server are performed with this class

    The requests.Response object will be extended to imitate a flask Response
    object by adding data and mimetype.
    """

    # The default server is the localhost
    actinia_server = "http://127.0.0.1:5000"
    # Check for an environmental variable to set the hostname http://IP:PORT
    if "ACTINIA_SERVER" in os.environ:
        actinia_server = str(os.environ["ACTINIA_SERVER"])

    def _make_flask_response(self, resp):
        """Take care of the mimetype to avoid r.text to take forever to parse
        a binary file.

        Args:
            resp: The requests.Response object that needs to be extended to be
                  a flask Response object

        Returns: extended requests.Response object with mimetype and data
                 attributes

        """
        if "content-type" in resp.headers:
            resp.mimetype = resp.headers["content-type"].split(";")[0]
        if (
            "content-type" in resp.headers
            and "tif" in resp.headers["content-type"]
        ):
            resp.data = resp.raw
        elif (
            "content-type" in resp.headers
            and "png" in resp.headers["content-type"]
        ):
            resp.data = resp.raw
        else:
            resp.data = resp.text
        return resp

    def _request(self, url, method, **kargs):
        if "content_type" in kargs:
            del kargs["content_type"]

        if "http" not in url:
            server_url = self.actinia_server + url
        else:
            server_url = url

        if method == "post":
            resp = requests.post(server_url, **kargs)
        elif method == "get":
            if ".tif" in url:
                resp = requests.get(server_url, stream=True, **kargs)
            else:
                resp = requests.get(server_url, **kargs)
        elif method == "delete":
            resp = requests.delete(server_url, **kargs)
        else:
            resp = requests.put(server_url, **kargs)
        return self._make_flask_response(resp)

    def post(self, url, **kargs):
        return self._request(url=url, method="post", **kargs)

    def get(self, url, **kargs):
        return self._request(url=url, method="get", **kargs)

    def delete(self, url, **kargs):
        return self._request(url=url, method="delete", **kargs)

    def put(self, url, **kargs):
        return self._request(url=url, method="put", **kargs)


class ActiniaTestCaseBase(unittest.TestCase):
    """Base class for GRASS GIS REST API tests"""

    server_test = False
    custom_actinia_cfg = False
    guest = None
    user = None
    admin = None
    root = None
    auth_header = {}
    users_list = []

    if "ACTINIA_SERVER_TEST" in os.environ:
        server_test = bool(os.environ["ACTINIA_SERVER_TEST"])

    if "ACTINIA_CUSTOM_TEST_CFG" in os.environ:
        custom_actinia_cfg = str(os.environ["ACTINIA_CUSTOM_TEST_CFG"])

    @classmethod
    def setUpClass(cls):
        if cls.server_test is False and cls.custom_actinia_cfg is False:
            global_config.REDIS_SERVER_SERVER = "localhost"
            global_config.REDIS_SERVER_PORT = 7000
            global_config.GRASS_RESOURCE_DIR = "/tmp"
            global_config.DOWNLOAD_CACHE = "/tmp/download_cache"
            global_config.REDIS_QUEUE_SERVER_URL = "localhost"
            global_config.REDIS_QUEUE_SERVER_PORT = 6379
            global_config.NUMBER_OF_WORKERS = 3
            # Create the job queue
            # redis_interface.create_job_queues(
            #     global_config.REDIS_QUEUE_SERVER_URL,
            #     global_config.REDIS_QUEUE_SERVER_PORT,
            #     global_config.NUMBER_OF_WORKERS)

        # If the custom_actinia_cfg variable is set, then the actinia config
        # file will be read to configure Redis queue
        if cls.server_test is False and cls.custom_actinia_cfg is not False:
            global_config.read(cls.custom_actinia_cfg)

            # # Create the job queue
            # redis_interface.create_job_queues(
            #     global_config.REDIS_QUEUE_SERVER_URL,
            #     global_config.REDIS_QUEUE_SERVER_PORT,
            #     global_config.NUMBER_OF_WORKERS
            # )

        # create configured folders
        os.makedirs(global_config.GRASS_USER_DATABASE, exist_ok=True)
        os.makedirs(global_config.GRASS_TMP_DATABASE, exist_ok=True)
        os.makedirs(global_config.GRASS_RESOURCE_DIR, exist_ok=True)
        os.makedirs(global_config.TMP_WORKDIR, exist_ok=True)
        os.makedirs(global_config.DOWNLOAD_CACHE, exist_ok=True)

        # Start the redis interface
        redis_args = (
            global_config.REDIS_SERVER_URL,
            global_config.REDIS_SERVER_PORT,
        )
        if (
            global_config.REDIS_SERVER_PW
            and global_config.REDIS_SERVER_PW is not None
        ):
            redis_args = (*redis_args, global_config.REDIS_SERVER_PW)

        redis_interface.connect(*redis_args)

        # Process queue
        create_process_queue(config=global_config)

        # We create 4 user for all roles: guest, user, admin, root
        accessible_datasets = {
            "nc_spm_08": [
                "PERMANENT",
                "user1",
                "landsat",
                "modis_lst",
                "test_mapset",
            ],
            "ECAD": ["PERMANENT"],
            "latlong_wgs84": ["PERMANENT"],
        }

        # Create users
        cls.guest_id, cls.guest_group, cls.guest_auth_header = cls.create_user(
            name="guest",
            role="guest",
            process_num_limit=3,
            process_time_limit=2,
            accessible_datasets=accessible_datasets,
        )
        cls.user_id, cls.user_group, cls.user_auth_header = cls.create_user(
            name="user",
            role="user",
            process_num_limit=3,
            process_time_limit=4,
            accessible_datasets=accessible_datasets,
        )
        cls.admin_id, cls.admin_group, cls.admin_auth_header = cls.create_user(
            name="admin", role="admin", accessible_datasets=accessible_datasets
        )
        cls.root_id, cls.root_group, cls.root_auth_header = cls.create_user(
            name="superadmin",
            role="superadmin",
            accessible_datasets=accessible_datasets,
        )

    @classmethod
    def create_user(
        cls,
        name="guest",
        role="guest",
        group="group",
        password="abcdefgh",
        accessible_datasets=None,
        process_num_limit=1000,
        process_time_limit=6000,
    ):
        auth = bytes("%s:%s" % (name, password), "utf-8")

        # We need to create an HTML basic authorization header
        cls.auth_header[role] = Headers()
        cls.auth_header[role].add(
            "Authorization", "Basic " + base64.b64encode(auth).decode()
        )

        # Make sure the user database is empty
        user = ActiniaUser(name)
        if user.exists():
            user.delete()
        # Create a user in the database
        user = ActiniaUser.create_user(
            name,
            group,
            password,
            user_role=role,
            accessible_datasets=accessible_datasets,
            process_num_limit=process_num_limit,
            process_time_limit=process_time_limit,
        )
        user.add_accessible_modules(["uname", "sleep"])
        cls.users_list.append(user)

        return name, group, cls.auth_header[role]

    @classmethod
    def tearDownClass(cls):
        for user in cls.users_list:
            user.delete()

        if cls.server_test is False:
            redis_interface.disconnect()

    def setUp(self):
        # We need to set the application context
        self.app_context = flask_app.app_context()
        self.app_context.push()

        # Check if the local or server site tests should be performed
        if self.server_test is False:
            flask_app.config["TESTING"] = True

            self.server = flask_app.test_client()
        else:
            self.server = ActiniaRequests()

    def tearDown(self):
        self.app_context.pop()

    def waitAsyncStatusAssertHTTP(
        self,
        response,
        headers,
        http_status=200,
        status="finished",
        message_check=None,
    ):
        """Poll the status of a resource and assert its finished HTTP status

        The response will be checked if the resource was accepted. Hence it
        must always be HTTP 200 status.

        The status URL from the response is then polled until status: finished,
        error or terminated.
        The result of the poll can be checked against its HTTP status and its
        actinia status message.

        Args:
            response: The accept response
            http_status (int): The HTTP status that should be checked
            status (str): The return status of the response
            message_check (str): A string that must be in the message field

        Returns: response

        """
        # Check if the resource was accepted
        self.assertEqual(
            response.status_code,
            200,
            "HTML status code is wrong %i" % response.status_code,
        )
        self.assertEqual(
            response.mimetype,
            "application/json",
            "Wrong mimetype %s" % response.mimetype,
        )

        resp_data = json_loads(response.data)

        rv_user_id = resp_data["user_id"]
        rv_resource_id = resp_data["resource_id"]

        while True:
            rv = self.server.get(
                URL_PREFIX + "/resources/%s/%s" % (rv_user_id, rv_resource_id),
                headers=headers,
            )
            resp_data = json_loads(rv.data)
            if (
                resp_data["status"] == "finished"
                or resp_data["status"] == "error"
                or resp_data["status"] == "terminated"
                or resp_data["status"] == "timeout"
            ):
                break
            time.sleep(0.2)

        self.assertEqual(resp_data["status"], status)
        self.assertEqual(
            rv.status_code,
            http_status,
            "HTML status code is wrong %i" % rv.status_code,
        )

        if message_check is not None:
            self.assertTrue(
                message_check in resp_data["message"],
                (f"Message is {resp_data['message']}"),
            )

        time.sleep(0.4)
        return resp_data

    def assertRasterInfo(self, location, mapset, raster, ref_info, header):
        url = (
            f"{URL_PREFIX}/locations/{location}/mapsets/{mapset}/"
            f"raster_layers/{raster}"
        )
        rv = self.server.get(url, headers=header)
        resp = json_loads(rv.data.decode())
        info = resp["process_results"]
        for key, val in ref_info.items():
            self.assertIn(
                key, info, f"RasterInfoAssertion failed: key {key} not found"
            )
            self.assertEqual(
                val,
                info[key],
                (
                    f"RasterInfoAssertion failed:"
                    f" value {key}:{val} does not match reference"
                ),
            )

    def assertVectorInfo(self, location, mapset, vector, ref_info, header):
        url = (
            f"{URL_PREFIX}/locations/{location}/mapsets/{mapset}/"
            f"vector_layers/{vector}"
        )
        rv = self.server.get(url, headers=header)
        resp = json_loads(rv.data.decode())
        info = resp["process_results"]
        for key, val in ref_info.items():
            self.assertIn(
                key, info, f"VectorInfoAssertion failed: key {key} not found"
            )
            self.assertEqual(
                val,
                info[key],
                (
                    f"VectorInfoAssertion failed:"
                    f" value {key}:{val} does not match reference"
                ),
            )

    def create_new_mapset(self, mapset_name, location_name="nc_spm_08"):
        self.delete_mapset(mapset_name, location_name)
        # Create new mapset
        self.server.post(
            URL_PREFIX
            + "/locations/%s/mapsets/%s" % (location_name, mapset_name),
            headers=self.admin_auth_header,
        )

    def delete_mapset(self, mapset_name, location_name="nc_spm_08"):
        # Unlock mapset for deletion
        self.server.delete(
            URL_PREFIX
            + "/locations/%s/mapsets/%s/lock" % (location_name, mapset_name),
            headers=self.admin_auth_header,
        )

        # Delete existing mapset
        self.server.delete(
            URL_PREFIX
            + "/locations/%s/mapsets/%s" % (location_name, mapset_name),
            headers=self.admin_auth_header,
        )

    def create_vector_layer(self, location, mapset, vector, region, parameter):
        # Remove potentially existing vector layer
        url = (
            f"{URL_PREFIX}/locations/{location}/mapsets/{mapset}/"
            f"vector_layers/{vector}"
        )
        rv = self.server.delete(url, headers=self.user_auth_header)

        parameter["column"] = "z"
        region["res"] = 100000
        # Create processing chain for random vector creation
        postbody = {
            "list": [
                {
                    "id": "set_region",
                    "module": "g.region",
                    "inputs": [
                        {"param": key, "value": str(val)}
                        for key, val in region.items()
                    ],
                },
                {
                    "id": "create_vector",
                    "module": "v.random",
                    "inputs": [
                        {"param": key, "value": str(val)}
                        for key, val in parameter.items()
                    ],
                    "outputs": [{"param": "output", "value": vector}],
                    "flags": "z",
                },
            ],
            "version": "1",
        }
        url = (
            f"{URL_PREFIX}/locations/{location}/mapsets/{mapset}/"
            f"processing_async"
        )
        rv = self.server.post(
            url,
            headers=self.user_auth_header,
            data=json_dumps(postbody),
            content_type="application/json",
        )
        self.waitAsyncStatusAssertHTTP(
            rv,
            headers=self.admin_auth_header,
            http_status=200,
            status="finished",
        )

        self.assertEqual(
            rv.status_code,
            200,
            "HTML status code is wrong %i" % rv.status_code,
        )
        self.assertEqual(
            rv.mimetype, "application/json", "Wrong mimetype %s" % rv.mimetype
        )



# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2016-2022 Sören Gebbert and mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""
Raster layer resources
"""

from actinia_api.swagger2.actinia_core.schemas.raster_layer import (
    RasterInfoModel,
)
from actinia_core.processing.actinia_processing.ephemeral_processing import (
    EphemeralProcessing,
)

__license__ = "GPLv3"
__author__ = "Sören Gebbert, Carmen Tawalika, Guido Riembauer, Anika Weinmann"
__copyright__ = (
    "Copyright 2016-2022, Sören Gebbert and mundialis GmbH & Co. KG"
)
__maintainer__ = "mundialis"


class EphemeralRasterInfo(EphemeralProcessing):
    def __init__(self, *args):
        EphemeralProcessing.__init__(self, *args)

    def _execute(self):
        """Read info from a raster layer

        Use a temporary mapset for processing
        """
        self._setup()

        raster_name = self.map_name
        self.required_mapsets.append(self.mapset_name)

        pc = {}
        pc["1"] = {
            "module": "r.info",
            "inputs": {"map": raster_name + "@" + self.mapset_name},
            "flags": "gre",
        }

        self.skip_region_check = True
        process_list = (
            self._create_temporary_grass_environment_and_process_list(
                process_chain=pc, skip_permission_check=True
            )
        )
        self._execute_process_list(process_list)

        kv_list = self.module_output_log[0]["stdout"].split("\n")

        raster_info = {}

        for string in kv_list:
            if "=" in string:
                k, v = string.split("=", 1)
                raster_info[k] = v

        self.module_results = RasterInfoModel(**raster_info)



# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2021-2022 mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""
This code shows the STACExporter model which help actinia core
to export and manage the STAC items a global collection configuration

"""

__license__ = "GPLv3"
__author__ = "Jorge Herrera"
__copyright__ = "Copyright 2021-2022, mundialis GmbH & Co. KG"
__maintainer__ = "__mundialis__"
__email__ = "info@mundialis.de"

from datetime import datetime
import numpy as np
import pyproj
from pystac import Item, read_dict, Catalog, Asset

from shapely.ops import transform
import rasterio
from shapely.geometry import Polygon, mapping

from actinia_core.core.common.exceptions import AsyncProcessTermination
from actinia_core.core.common.app import API_VERSION
from actinia_core.version import G_VERSION

try:
    from actinia_stac_plugin.core.common import connectRedis
    from actinia_stac_plugin.core.stac_redis_interface import (
        redis_actinia_interface,
    )
    from pystac.extensions.projection import (
        ItemProjectionExtension as ProjectionItemExt,
    )

    has_plugin = True
except Exception:
    has_plugin = False


class STACExporter:
    def stac_builder(self, resource_url: str, filename: str, output_type: str):
        """
        This function build the STAC ITEM and implement the following
        extension:
            - Projection
            - Raster
            Args:
                - output_path = Path to the source
                - filename =  name of the source
                - output_type =  type of object (raster, vector)
        """

        output_path = self._get_source_file(resource_url)

        self._stac_collection_initializer()

        if output_type == "raster":
            # Get parameters for STAC item
            extra_values = self._get_raster_parameters(output_path)

            # Checking if the input has WGS 84 CRS
            geom, bbox_raster = self._get_wgs84_parameters(extra_values)

            item_name = f"item-{filename}"

            # Start building the Item
            item = Item(
                id=item_name,
                geometry=geom,
                bbox=bbox_raster,
                datetime=datetime.utcnow(),
                properties={"gds": extra_values["gds"]},
            )

            # Adding Asset and Raster Ext
            asset_ = Asset(href=resource_url, title=filename)

            item.add_asset(key="source", asset=asset_)

            # Adding the Projection Extension
            proj_ext = ProjectionItemExt(item)
            proj_ext.apply(
                epsg=extra_values["crs"],
                geometry=extra_values["geometry"],
                shape=extra_values["shape"],
                bbox=extra_values["bbox"],
                transform=extra_values["transform"],
            )

            # Adding the Raster Extension
            item = self._set_raster_extention(output_path, item)

            # Adding the Processing Extension
            item = self._set_processing_extention(item)

            # Read catalog from REDIS
            catalog_dict = redis_actinia_interface.read("result-catalog")

            # Create Catalog Object
            catalog = read_dict(catalog_dict)

            redis_items = self._save_items_redis(item, item_name)

            if redis_items:
                # Add item to Catalog
                catalog.add_item(item)

                # Update redis catalog
                self._update_catalog_redis(catalog)

                return item.to_dict()

        # TODO
        elif output_type == "vector":
            raise AsyncProcessTermination("Not implemented yet.")

        return item.to_dict()

    @staticmethod
    def _get_source_file(url):
        uri_split = url.split(API_VERSION)

        source_file = f"/actinia_core{uri_split[1]}"

        return source_file

    @staticmethod
    def _stac_collection_initializer():
        """
        Initialize the STAC Catalog for the different outputs in actinia
        Catalog allows to have versability on the spatio-temporal spectrum,
        in addition the properties are independent for each item stored

        Code uses pystac as base for the creation of stac catalogs
        """
        connectRedis()

        result_catalog_validation = redis_actinia_interface.exists(
            "result-catalog"
        )

        if not result_catalog_validation:
            results = Catalog(id="result-catalog", description="STAC catalog")
            results.normalize_and_save("/stac/catalogs")

            redis_actinia_interface.create(
                "result-catalog",
                results.to_dict(),
            )
            return "result-catalog was created"
        else:
            return "result-catalog is already created"

    @staticmethod
    def _save_items_redis(item, item_name):
        try:
            connectRedis()

            exist = redis_actinia_interface.exists(item_name)

            if exist:
                redis_actinia_interface.update(item_name, item.to_dict())
                return False

            redis_actinia_interface.create(
                item_name,
                item.to_dict(),
            )
            return True
        except AssertionError:
            raise AssertionError("Something went wrong")

    @staticmethod
    def _get_raster_parameters(raster_path):
        with rasterio.open(raster_path) as raster:
            gds = raster.transform[:]
            bounds = raster.bounds
            bbox = [bounds.left, bounds.bottom, bounds.right, bounds.top]
            geom = Polygon(
                [
                    [bounds.left, bounds.bottom],
                    [bounds.left, bounds.top],
                    [bounds.right, bounds.top],
                    [bounds.right, bounds.bottom],
                ]
            )
            crs = raster.crs.to_epsg()

            extra_values = {
                "gds": gds[0],
                "crs": crs,
                "bbox": bbox,
                "geometry": mapping(geom),
                "transform": gds,
                "shape": raster.shape,
                "datetime": "2021-12-09T16:41:39.985257Z",
            }

            return extra_values

    @staticmethod
    def _get_wgs84_parameters(extra_values):
        if extra_values["crs"] != 4326:
            wgs84 = pyproj.CRS("EPSG:4326")
            raster_proj = pyproj.CRS("EPSG:" + str(extra_values["crs"]))
            project = pyproj.Transformer.from_crs(
                raster_proj, wgs84, always_xy=True
            ).transform
            geojson = Polygon(
                [tuple(i) for i in extra_values["geometry"]["coordinates"][0]]
            )
            geom = transform(project, geojson)
            geom = mapping(geom)

            def bbox(coord_list):
                box = []
                for i in (0, 1):
                    res = sorted(coord_list, key=lambda x: x[i])
                    box.append((res[0][i], res[-1][i]))
                bbox = [box[0][0], box[1][0], box[0][1], box[1][1]]
                return bbox

            bbox_raster = bbox(list(geom["coordinates"]))
        else:
            geom = extra_values["geometry"]
            bbox_raster = extra_values["bbox"]

        return geom, bbox_raster

    @staticmethod
    def _update_catalog_redis(catalog):
        new_catalog = catalog.to_dict()
        redis_actinia_interface.update("result-catalog", new_catalog)

    @staticmethod
    def _set_processing_extention(item):
        input_item = item.to_dict()

        input_item["processing:facility"] = (f"Actinia Core {API_VERSION}",)
        input_item["processing:level"] = "L4"
        input_item["processing:derived_from"] = "https://actinia.mundialis.de/"
        input_item["processing:software"] = f" GRASS {G_VERSION}"
        proc_schema = (
            "https://stac-extensions.github.io/processing/v1.1.0/schema.json"
        )

        input_item["stac_extensions"].append(proc_schema)

        proc_ext_item = read_dict(input_item)

        return proc_ext_item

    @staticmethod
    def _set_raster_extention(raster_path, item):
        with rasterio.open(raster_path) as raster:
            band = raster.read(1)
            pixelSizeX, pixelSizeY = raster.res

            nodata = np.count_nonzero(np.isnan(band))
            spatial_resolution = pixelSizeX
            data_type = str(band.dtype)

            input_item = item.to_dict()
            asset = input_item["assets"]["source"]
            asset["raster:nodata"] = nodata
            asset["raster:spatial_resolution"] = spatial_resolution
            asset["raster:data_type"] = data_type
            proc_schema = (
                "https://stac-extensions.github.io/raster/v1.1.0/schema.json"
            )
            input_item["stac_extensions"].append(proc_schema)

            ras_ext_item = read_dict(input_item)

        return ras_ext_item



# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2016-2022 Sören Gebbert and mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""
Mapset management resources

* List all mapsets
* Create mapset, Delete mapset, Get info about a mapset
* Lock mapset, unlock mapset, get mapset lock status
"""

from flask import jsonify, make_response
from flask_restful_swagger_2 import swagger
import pickle
from actinia_api.swagger2.actinia_core.apidocs import mapset_management

from actinia_core.rest.base.resource_base import ResourceBase
from actinia_core.core.common.app import auth
from actinia_core.core.common.api_logger import log_api_call
from actinia_core.core.common.redis_interface import enqueue_job
from actinia_core.rest.base.endpoint_config import (
    check_endpoint,
    endpoint_decorator,
)
from actinia_core.rest.base.user_auth import check_user_permissions
from actinia_core.rest.base.user_auth import check_admin_role, check_user_role
from actinia_core.processing.common.mapset_management import (
    list_raster_mapsets,
    read_current_region,
    create_mapset,
    delete_mapset,
    get_mapset_lock,
    lock_mapset,
    unlock_mapset,
)

__license__ = "GPLv3"
__author__ = "Sören Gebbert, Carmen Tawalika, Julia Haas"
__copyright__ = (
    "Copyright 2016-2022, Sören Gebbert and mundialis GmbH & Co. KG"
)
__maintainer__ = "mundialis"


class ListMapsetsResource(ResourceBase):
    """List all mapsets in a location"""

    layer_type = None

    # @check_queue_type_overwrite()
    @endpoint_decorator()
    @swagger.doc(check_endpoint("get", mapset_management.get_doc))
    def get(self, location_name):
        """
        Get a list of all mapsets that are located in a specific location.
        """
        rdc = self.preprocess(
            has_json=False,
            has_xml=False,
            location_name=location_name,
            mapset_name="PERMANENT",
        )
        if rdc:
            enqueue_job(
                self.job_timeout,
                list_raster_mapsets,
                rdc,
                queue_type_overwrite=True,
            )
            http_code, response_model = self.wait_until_finish()
        else:
            http_code, response_model = pickle.loads(self.response_data)

        return make_response(jsonify(response_model), http_code)


class MapsetManagementResourceUser(ResourceBase):
    """This class returns information about a mapset"""

    def __init__(self):
        ResourceBase.__init__(self)

    @endpoint_decorator()
    @swagger.doc(check_endpoint("get", mapset_management.get_user_doc))
    def get(self, location_name, mapset_name):
        """
        Get the current computational region of the mapset and the projection
        of the location as WKT string.
        """
        rdc = self.preprocess(
            has_json=False,
            has_xml=False,
            location_name=location_name,
            mapset_name=mapset_name,
        )

        enqueue_job(
            self.job_timeout,
            read_current_region,
            rdc,
            queue_type_overwrite=True,
        )
        http_code, response_model = self.wait_until_finish()
        return make_response(jsonify(response_model), http_code)


class MapsetManagementResourceAdmin(ResourceBase):
    """This class manages the creation, deletion and modification of mapsets

    This is allowed for administrators and users
    """

    decorators = [
        log_api_call,
        check_user_permissions,
        check_user_role,
        auth.login_required,
    ]

    def __init__(self):
        ResourceBase.__init__(self)

    @endpoint_decorator()
    @swagger.doc(check_endpoint("post", mapset_management.post_user_doc))
    def post(self, location_name, mapset_name):
        """Create a new mapset in an existing location."""
        rdc = self.preprocess(
            has_json=False,
            has_xml=False,
            location_name=location_name,
            mapset_name=mapset_name,
        )

        enqueue_job(self.job_timeout, create_mapset, rdc)
        http_code, response_model = self.wait_until_finish()
        return make_response(jsonify(response_model), http_code)

    def put(self, location_name, mapset_name):
        """Modify the region of a mapset

        TODO: Implement region setting

        Args:
            location_name (str): Name of the location
            mapset_name (str): Name of the mapset

        Returns:
            flaks.Response:
            HTTP 200 and JSON document in case of success, HTTP 400 otherwise

        """
        pass

    @endpoint_decorator()
    @swagger.doc(check_endpoint("delete", mapset_management.delete_user_doc))
    def delete(self, location_name, mapset_name):
        """Delete an existing mapset"""
        rdc = self.preprocess(
            has_json=False,
            has_xml=False,
            location_name=location_name,
            mapset_name=mapset_name,
        )

        enqueue_job(
            self.job_timeout, delete_mapset, rdc, queue_type_overwrite=True
        )
        http_code, response_model = self.wait_until_finish()
        return make_response(jsonify(response_model), http_code)


class MapsetLockManagementResource(ResourceBase):
    """Lock a mapset"""

    decorators = [
        log_api_call,
        check_user_permissions,
        check_admin_role,
        auth.login_required,
    ]

    @endpoint_decorator()
    @swagger.doc(check_endpoint("get", mapset_management.get_lock_doc))
    def get(self, location_name, mapset_name):
        """Get the location/mapset lock status."""
        rdc = self.preprocess(
            has_json=False,
            has_xml=False,
            location_name=location_name,
            mapset_name=mapset_name,
        )

        enqueue_job(
            self.job_timeout, get_mapset_lock, rdc, queue_type_overwrite=True
        )
        http_code, response_model = self.wait_until_finish()
        return make_response(jsonify(response_model), http_code)

    @endpoint_decorator()
    @swagger.doc(check_endpoint("post", mapset_management.post_lock_doc))
    def post(self, location_name, mapset_name):
        """Create a location/mapset lock."""
        rdc = self.preprocess(
            has_json=False,
            has_xml=False,
            location_name=location_name,
            mapset_name=mapset_name,
        )

        enqueue_job(self.job_timeout, lock_mapset, rdc)
        http_code, response_model = self.wait_until_finish()
        return make_response(jsonify(response_model), http_code)

    @endpoint_decorator()
    @swagger.doc(check_endpoint("delete", mapset_management.delete_lock_doc))
    def delete(self, location_name, mapset_name):
        """Delete a location/mapset lock."""
        rdc = self.preprocess(
            has_json=False,
            has_xml=False,
            location_name=location_name,
            mapset_name=mapset_name,
        )

        enqueue_job(
            self.job_timeout, unlock_mapset, rdc, queue_type_overwrite=True
        )
        http_code, response_model = self.wait_until_finish()
        return make_response(jsonify(response_model), http_code)



# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2016-2018 Sören Gebbert and mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""
Tests: Location test case
"""
from flask.json import loads as json_loads, dumps as json_dumps
import unittest

try:
    from .test_resource_base import ActiniaResourceTestCaseBase, URL_PREFIX
except ModuleNotFoundError:
    from test_resource_base import ActiniaResourceTestCaseBase, URL_PREFIX

__license__ = "GPLv3"
__author__ = "Sören Gebbert, Anika Weinmann"
__copyright__ = (
    "Copyright 2016-2022, Sören Gebbert and mundialis GmbH & Co. KG"
)
__maintainer__ = "Sören Gebbert"
__email__ = "soerengebbert@googlemail.com"


class LocationTestCase(ActiniaResourceTestCaseBase):
    def test_list_locations(self):
        rv = self.server.get(
            URL_PREFIX + "/locations", headers=self.user_auth_header
        )
        print(rv.data)
        self.assertEqual(
            rv.status_code,
            200,
            "HTML status code is wrong %i" % rv.status_code,
        )
        self.assertEqual(
            rv.mimetype, "application/json", "Wrong mimetype %s" % rv.mimetype
        )

        if "nc_spm_08" in json_loads(rv.data)["locations"]:
            location = "nc_spm_08"

        self.assertEqual(location, "nc_spm_08", "Wrong location listed")

    def test_location_info(self):
        rv = self.server.get(
            URL_PREFIX + "/locations/nc_spm_08/info",
            headers=self.admin_auth_header,
        )
        print(rv.data)
        self.assertEqual(
            rv.status_code,
            200,
            "HTML status code is wrong %i" % rv.status_code,
        )
        self.assertEqual(
            rv.mimetype, "application/json", "Wrong mimetype %s" % rv.mimetype
        )

        region_settings = json_loads(rv.data)["process_results"]["region"]
        json_loads(rv.data)["process_results"]["projection"]

        self.assertTrue("depths" in region_settings)
        self.assertTrue("ewres" in region_settings)
        self.assertTrue("cols" in region_settings)
        self.assertTrue("rows" in region_settings)

    def test_location_global_db_error(self):
        # ERROR: Try to create a location as admin that exists in the global
        # database
        rv = self.server.post(
            URL_PREFIX + "/locations/nc_spm_08",
            data=json_dumps({"epsg": "4326"}),
            content_type="application/json",
            headers=self.admin_auth_header,
        )
        print(rv.data)
        self.assertEqual(
            rv.status_code,
            400,
            "HTML status code is wrong %i" % rv.status_code,
        )
        self.assertEqual(
            rv.mimetype, "application/json", "Wrong mimetype %s" % rv.mimetype
        )

    def test_location_creation_and_deletion(self):
        # Delete a potentially existing location
        rv = self.server.delete(
            URL_PREFIX + "/locations/test_location",
            headers=self.admin_auth_header,
        )

        # Create new location as admin
        rv = self.server.post(
            URL_PREFIX + "/locations/test_location",
            data=json_dumps({"epsg": "4326"}),
            content_type="application/json",
            headers=self.admin_auth_header,
        )
        print(rv.data)
        self.assertEqual(
            rv.status_code,
            200,
            "HTML status code is wrong %i" % rv.status_code,
        )
        self.assertEqual(
            rv.mimetype, "application/json", "Wrong mimetype %s" % rv.mimetype
        )

        # ERROR: Try to create a location as admin that already exists
        rv = self.server.post(
            URL_PREFIX + "/locations/test_location",
            data=json_dumps({"epsg": "4326"}),
            content_type="application/json",
            headers=self.admin_auth_header,
        )
        print(rv.data)
        self.assertEqual(
            rv.status_code,
            400,
            "HTML status code is wrong %i" % rv.status_code,
        )
        self.assertEqual(
            rv.mimetype, "application/json", "Wrong mimetype %s" % rv.mimetype
        )

        # Delete location
        rv = self.server.delete(
            URL_PREFIX + "/locations/test_location",
            headers=self.admin_auth_header,
        )
        print(rv.data)
        self.assertEqual(
            rv.status_code,
            200,
            "HTML status code is wrong %i" % rv.status_code,
        )
        self.assertEqual(
            rv.mimetype, "application/json", "Wrong mimetype %s" % rv.mimetype
        )

        # ERROR: Delete should fail, since location does not exists
        rv = self.server.delete(
            URL_PREFIX + "/locations/test_location",
            headers=self.admin_auth_header,
        )
        print(rv.data)
        self.assertEqual(
            rv.status_code,
            400,
            "HTML status code is wrong %i" % rv.status_code,
        )
        self.assertEqual(
            rv.mimetype, "application/json", "Wrong mimetype %s" % rv.mimetype
        )

    def test_location_creation_and_deletion_as_user(self):
        # Delete a potentially existing location
        rv = self.server.delete(
            URL_PREFIX + "/locations/test_location",
            headers=self.user_auth_header,
        )

        # Create new location as user
        rv = self.server.post(
            URL_PREFIX + "/locations/test_location",
            data=json_dumps({"epsg": "4326"}),
            content_type="application/json",
            headers=self.user_auth_header,
        )
        self.assertEqual(
            rv.status_code,
            200,
            "Location creation by user: HTML status code is wrong %i"
            % rv.status_code,
        )
        self.assertEqual(
            rv.mimetype,
            "application/json",
            "Location creation by user: Wrong mimetype %s" % rv.mimetype,
        )

        # ERROR: Try to create a location as user that already exists
        rv = self.server.post(
            URL_PREFIX + "/locations/test_location",
            data=json_dumps({"epsg": "4326"}),
            content_type="application/json",
            headers=self.user_auth_header,
        )
        self.assertEqual(
            rv.status_code,
            400,
            "Location recreation by user: HTML status code is wrong %i"
            % rv.status_code,
        )
        self.assertEqual(
            rv.mimetype,
            "application/json",
            "Location recreation by user: Wrong mimetype %s" % rv.mimetype,
        )

        # Delete location
        rv = self.server.delete(
            URL_PREFIX + "/locations/test_location",
            headers=self.user_auth_header,
        )
        self.assertEqual(
            rv.status_code,
            200,
            "Location deletion by user: HTML status code is wrong %i"
            % rv.status_code,
        )
        self.assertEqual(
            rv.mimetype,
            "application/json",
            "Location deletion by user: Wrong mimetype %s" % rv.mimetype,
        )

        # ERROR: Delete should fail, since location does not exists
        rv = self.server.delete(
            URL_PREFIX + "/locations/test_location",
            headers=self.user_auth_header,
        )
        self.assertEqual(
            rv.status_code,
            400,
            "Location redeletion by user: HTML status code is wrong %i"
            % rv.status_code,
        )
        self.assertEqual(
            rv.mimetype,
            "application/json",
            "Location redeletion by user: Wrong mimetype %s" % rv.mimetype,
        )

    def test_location_creation_and_deletion_as_guest(self):
        # ERROR: Try to create a location as guest
        rv = self.server.post(
            URL_PREFIX + "/locations/test_location_user",
            data=json_dumps({"epsg": "4326"}),
            content_type="application/json",
            headers=self.guest_auth_header,
        )
        print(rv.data)
        self.assertEqual(
            rv.status_code,
            401,
            "HTML status code is wrong %i" % rv.status_code,
        )
        self.assertEqual(
            rv.mimetype, "application/json", "Wrong mimetype %s" % rv.mimetype
        )

        # ERROR: Delete should fail since the guest user is not authorized
        rv = self.server.delete(
            URL_PREFIX + "/locations/test_location_user",
            headers=self.guest_auth_header,
        )
        print(rv.data)
        self.assertEqual(
            rv.status_code,
            401,
            "HTML status code is wrong %i" % rv.status_code,
        )
        self.assertEqual(
            rv.mimetype, "application/json", "Wrong mimetype %s" % rv.mimetype
        )


if __name__ == "__main__":
    unittest.main()
